#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.conf import settings
from django.core.urlresolvers import reverse
from django.db.models import Count, Q

import beat.core.algorithm
import beat.core.library

from beat.core.utils import NumpyJSONEncoder

from ..dataformats.models import DataFormat

from ..common.storage import OverwriteStorage
from ..common.exceptions import ShareError
from ..common.models import get_contribution_declaration_filename
from ..common.models import get_contribution_description_filename

from ..code.models import Code
from ..code.models import CodeManager
from ..code.models import get_contribution_source_code_filename
from ..libraries.models import Library

import simplejson
import collections
import itertools


#----------------------------------------------------------


def validate_algorithm(declaration):
    """Validates the declaration of an algorithm code, returns wrapper"""

    if not declaration:
        raise SyntaxError('Algorithm declaration cannot be empty')

    if not(isinstance(declaration, dict)):
        try:
            declaration_dict = simplejson.loads(declaration, object_pairs_hook=collections.OrderedDict)
        except Exception as e:
            raise SyntaxError(str(e))
    else:
        declaration_dict = collections.OrderedDict(declaration)

    algorithm = beat.core.algorithm.Algorithm(settings.PREFIX, declaration_dict)

    if not algorithm.valid:
        raise SyntaxError('\n  * %s' % '\n  * '.join(algorithm.errors))

    return algorithm


#----------------------------------------------------------


class AlgorithmStorage(OverwriteStorage):

    def __init__(self, *args, **kwargs):
        kwargs['location'] = settings.ALGORITHMS_ROOT
        super(AlgorithmStorage, self).__init__(*args, **kwargs)


#----------------------------------------------------------


class AlgorithmManager(CodeManager):

    def create_algorithm(self, author, name, short_description='',
                         description='', declaration=None, code=None, version=1,
                         previous_version=None, fork_of=None):

        default = beat.core.algorithm.Algorithm(settings.PREFIX, data=None)

        return self.create_code(author, name, default, short_description,
                                description, declaration, code, version,
                                previous_version, fork_of)


#----------------------------------------------------------


class Algorithm(Code):

    #_____ Fields __________

    declaration_file = models.FileField(
                                     storage=AlgorithmStorage(),
                                     upload_to=get_contribution_declaration_filename,
                                     blank=True, null=True,
                                     max_length=200,
                                     db_column='declaration'
                                    )

    description_file = models.FileField(
                                    storage=AlgorithmStorage(),
                                    upload_to=get_contribution_description_filename,
                                    blank=True, null=True,
                                    max_length=200,
                                    db_column='description'
                                   )

    source_code_file = models.FileField(
                                    storage=AlgorithmStorage(),
                                    upload_to=get_contribution_source_code_filename,
                                    blank=True, null=True,
                                    max_length=200,
                                    db_column='source_code'
                                   )

    # Read-only parameters that are updated at every save(), if required
    parameters        = models.TextField(blank=True, null=True)
    result_dataformat = models.TextField(blank=True, null=True)
    splittable        = models.BooleanField(default=False,
                                            help_text='Defines if the code can be executed ' \
                                                    'in multiple instances')

    referenced_libraries = models.ManyToManyField(Library,
                                                  blank=True, related_name='used_by_algorithms',
                                                  symmetrical=False)

    objects = AlgorithmManager()


    #_____ Utilities __________

    def get_absolute_url(self):

        return reverse(
                'algorithms:view',
                args=(self.author.username, self.name, self.version,),
               )


    def get_api_update_url(self):
        '''Returns the endpoint to update this object'''

        return reverse(
                'api_algorithms:object',
                args=(self.author.username, self.name, self.version,),
               )


    def get_api_share_url(self):
        '''Returns the endpoint to share this object'''

        return reverse(
                'api_algorithms:share',
                args=(self.author.username, self.name, self.version,),
               )


    def environments(self):
        '''Calculates environment usage for this algorithm

        Returns:

          list: mapping environment to usage counts, determining how many times
            a given algorithm has been successfuly used on that environment

        '''

        from ..experiments.models import Block
        from ..experiments.models import Experiment
        from ..backend.models import Environment


        # Tries to figure through a maximum if an algorithm has been
        # successfuly used inside an environment.

        # Case 1) The block is part of an experiment that was successful
        # Case 2) The block is part of an experiment that is not successful
        # (failed or other), but it is CACHED (if not cached, then we can't
        # attest anything about the algorithm/environment relationship!)

        envs = Environment.objects.filter(blocks__in=self.blocks.filter(
            Q(experiment__status=Experiment.DONE) | \
            ((~Q(experiment__status=Experiment.DONE)) & Q(status=Block.CACHED))
            )).annotate(itemcount=Count('id')).order_by('-creation_date').distinct()

        return [(k, k.itemcount) for k in envs]


    #_____ Overrides __________

    def save(self, *args, **kwargs):
        wrapper = self._save_preprocessing()

        if not wrapper.parameters:
            self.parameters = None
        else:
            parameters = []
            for name, details in wrapper.parameters.items():
                parameters.append({
                    "name":          name,
                    "default_value": details.get('default'),
                    "comment":       details.get('description', ''),
                    "type":          details['type'],
                    }
                )

                if 'choice' in details:
                    parameters[-1]['choices'] = details['choice']

                if 'range' in details:
                    parameters[-1]['minimum'] = details['range'][0]
                    parameters[-1]['maximum'] = details['range'][1]

            self.parameters = simplejson.dumps(parameters,
                                               indent=4,
                                               cls=NumpyJSONEncoder)

        if not wrapper.results:
            self.result_dataformat = None
        else:
            results = {}
            for name, details in wrapper.results.items():
                key = name if not details.get('display') else ('+' + name)
                results[key] = details['type']

            self.result_dataformat = simplejson.dumps(results,
                                                      indent=4,
                                                      cls=NumpyJSONEncoder)

        # Set splittability
        self.splittable = wrapper.splittable

        # Invoke the base implementation
        super(Algorithm, self).save(*args, **kwargs)

        # reset referenced libraries
        self.referenced_libraries.clear()
        if wrapper.uses is not None:
            for l in set(wrapper.uses.values()):
                s = beat.core.library.Storage(settings.PREFIX, l)
                library = Library.objects.get(author__username=s.username,
                                              name=s.name,
                                              version=s.version,
                                             )
                self.referenced_libraries.add(library)


    def share(self, public, users=None, teams=None):
        self._share_dataformats(users=users, teams=teams)
        super(Algorithm, self).share(public=public, users=users, teams=teams)


    #_____ Methods __________

    def validate(self, declaration):
        return validate_algorithm(declaration)

    def analysis(self):
        return (self.result_dataformat is not None)

    def all_referenced_result_dataformats(self):
        retval = set()
        if self.result_dataformat:
            result_dataformat = simplejson.loads(self.result_dataformat)
            for value in result_dataformat.values():
                elements = value.split('/')
                if len(elements) == 3:
                    retval.add(DataFormat.objects.get(author__username=elements[0], name=elements[1], version=elements[2]))
        return retval

    def all_referenced_dataformats(self):
        dataformats = set(map(lambda x: x.dataformat, self.endpoints.all()))
        return list(dataformats | self.all_referenced_result_dataformats())

    def all_needed_dataformats(self):
        result = self.all_referenced_dataformats()

        for dataformat in list(result):
            result.extend(dataformat.all_needed_dataformats())

        return list(set(result))

    def modifiable(self):
        return (self.experiments.count() == 0) and super(Algorithm, self).modifiable()

    def deletable(self):
        return (self.experiments.count() == 0) and super(Algorithm, self).deletable()


    def core(self):
        return validate_algorithm(self.declaration)


    def uses_and_groups(self):
        core = self.core()
        return core.uses, core.groups


    def json_parameters(self):
        return simplejson.loads(self.parameters) if self.parameters else []


    def json_result(self):
        return simplejson.loads(self.result_dataformat) if self.result_dataformat else {}


    #_____ Protected methods __________

    def _share_dataformats(self, users, teams):
        # Retrieve and process the list of referenced dataformats
        needed_formats = self.all_needed_dataformats()
        own_needed_formats = filter(lambda x: x.author == self.author, needed_formats)
        other_needed_formats = filter(lambda x: x.author != self.author, needed_formats)

        # Ensure that all needed dataformats from other users/teams have the necessary sharing
        # preferences
        errors = []
        for needed_format in other_needed_formats:
            errors.extend(needed_format.is_accessible(users=users, teams=teams))

        if len(errors) > 0:
            raise ShareError(errors)

        for dataformats in own_needed_formats:
            dataformats.share(users=users, teams=teams)


#----------------------------------------------------------


class AlgorithmEndpoint(models.Model):
    algorithm  = models.ForeignKey(Algorithm, related_name='endpoints')
    input      = models.BooleanField(default=False)
    name       = models.CharField(max_length=200)
    dataformat = models.ForeignKey(DataFormat, related_name='algorithm_endpoints')
    channel    = models.CharField(max_length=200)

    def __str__(self):
        return self.algorithm.fullname() + ', ' + self.name
