/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 * 
 * This file is part of the beat.web module of the BEAT platform.
 * 
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 * 
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.contributions.editor.js

   Declaration of the algorithm editor-related classes
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.contributions === undefined)
    beat.contributions = {}

if (beat.contributions.editor === undefined)
    beat.contributions.editor = {}



beat.contributions.editor.PROCESS_METHOD_ALGORITHM = 'def process(self, inputs, outputs):';
beat.contributions.editor.PROCESS_METHOD_ANALYZER  = 'def process(self, inputs, output):';
beat.contributions.editor.SETUP_METHOD_SIGNATURE   = 'def setup(self, parameters):';
beat.contributions.editor.SETUP_METHOD_CONTENT     = [ '# TODO: Use the parameters',
                                                    'return True'
                                                  ]


function name_check()
{
  this.value = this.value.trim().replace(/[^a-zA-Z0-9_\-]/g, '_');
}



/****************************** CLASS: SourceCodeEditor *********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor = function(id)
{
    this.editor = CodeMirror.fromTextArea($('#' + id)[0],
        {
            mode: 'python',
            foldGutter: true,
        }
    );
}


//----------------------------------------------------------------------------------------
// Returns the source code
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor.prototype.getSourceCode = function()
{
    return this.editor.getValue();
}


//----------------------------------------------------------------------------------------
// Add the setup() method in the source code (if necessary)
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor.prototype.addSetupMethod = function()
{
    // Retrieve the source code
    var lines          = this._getSourceCodeLines();
    var stripped_lines = this._stripLines(lines);

    // If the setup() method is already in the source code, don't do anything
    if (this._searchLine(stripped_lines, beat.contributions.editor.SETUP_METHOD_SIGNATURE))
        return;

    // Determine where to insert the new method
    var process_method_index = this._searchLine(stripped_lines, beat.contributions.editor.PROCESS_METHOD_ALGORITHM);

    if (process_method_index === null)
        process_method_index = this._searchLine(stripped_lines, beat.contributions.editor.PROCESS_METHOD_ANALYZER);

    if (process_method_index === null)
        process_method_index = lines.length;

    var indent = '    ';
    if (process_method_index < lines.length)
    {
        var line = lines[process_method_index];
        indent = line.substr(0, line.indexOf('def'));
    }

    // Generate the new source code
    var source_code = '';
    for (var i = 0; i < process_method_index; ++i)
        source_code += lines[i] + '\n';

    source_code += indent + beat.contributions.editor.SETUP_METHOD_SIGNATURE + '\n';
    for (var i = 0; i < beat.contributions.editor.SETUP_METHOD_CONTENT.length; ++i)
        source_code += indent + indent + beat.contributions.editor.SETUP_METHOD_CONTENT[i] + '\n';

    if (process_method_index < lines.length)
        source_code += '\n';

    for (var i = process_method_index; i < lines.length; ++i)
        source_code += lines[i] + '\n';

    this.editor.setValue(source_code);
}


//----------------------------------------------------------------------------------------
// Remove the setup() method in the source code (if not modified by the user)
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor.prototype.removeSetupMethod = function()
{
    // Retrieve the source code
    var lines          = this._getSourceCodeLines();
    var stripped_lines = this._stripLines(lines);

    // Search the position of the setup() method in the source code
    var index = this._searchLine(stripped_lines, beat.contributions.editor.SETUP_METHOD_SIGNATURE);
    if (index === null)
        return;

    // Determine if the method was modified by the user
    var stripped_content = this._stripLines(beat.contributions.editor.SETUP_METHOD_CONTENT);

    if (stripped_lines[index + 1] == stripped_content[1])
    {
        var nb = 2;
        while (stripped_lines[index + nb] == '')
            ++nb;

        lines.splice(index, nb);
    }

    else if (stripped_lines[index + 1] == stripped_content[0])
    {
        if (stripped_lines[index + 2] == stripped_content[1])
        {
            var nb = 3;
            while (stripped_lines[index + nb] == '')
                ++nb;

            lines.splice(index, nb);
        }
        else
        {
            return;
        }
    }

    else
    {
        return;
    }

    // Generate the new source code
    var source_code = '';
    for (var i = 0; i < lines.length; ++i)
        source_code += lines[i] + '\n';

    this.editor.setValue(source_code);
}


//----------------------------------------------------------------------------------------
// Modify the signature of the process() method in the source code
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor.prototype.changeProcessMethod = function(analyzer)
{
    // Retrieve the source code
    var lines          = this._getSourceCodeLines();
    var stripped_lines = this._stripLines(lines);

    // Search the position of the process() method in the source code
    var index = null;
    var new_signature = null;
    if (analyzer)
    {
        index = this._searchLine(stripped_lines, beat.contributions.editor.PROCESS_METHOD_ALGORITHM);
        new_signature = beat.contributions.editor.PROCESS_METHOD_ANALYZER;
    }
    else
    {
        index = this._searchLine(stripped_lines, beat.contributions.editor.PROCESS_METHOD_ANALYZER);
        new_signature = beat.contributions.editor.PROCESS_METHOD_ALGORITHM;
    }

    if (index === null)
        return;

    // Generate the new source code
    var indent = lines[index].substr(0, lines[index].indexOf('def'));

    var source_code = '';
    for (var i = 0; i < index; ++i)
        source_code += lines[i] + '\n';

    source_code += indent + new_signature + '\n';

    for (var i = index + 1; i < lines.length; ++i)
        source_code += lines[i] + '\n';

    this.editor.setValue(source_code);
}


//----------------------------------------------------------------------------------------
// Returns the source code splitted in lines
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor.prototype._getSourceCodeLines = function()
{
    var source_code = this.editor.getValue();

    var lines = source_code.split('\n');
    if (source_code[source_code.length-1] == '\n')
        lines.splice(lines.length - 1, 1);

    return lines;
}


//----------------------------------------------------------------------------------------
// Returns a stripped version of the provided source code lines
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor.prototype._stripLines = function(lines)
{
    return lines.map(function(x) { return x.trim().replace(/\s/g, ''); });
}


//----------------------------------------------------------------------------------------
// Returns the index of the provided line
//----------------------------------------------------------------------------------------
beat.contributions.editor.SourceCodeEditor.prototype._searchLine = function(stripped_lines, searched_line)
{
    var stripped = searched_line.trim().replace(/\s/g, '');

    for (var i = 0; i < stripped_lines.length; ++i)
    {
        if (stripped_lines[i] == stripped)
            return i;
    }

    return null;
}



/******************************** CLASS: InputsEditor ***********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor = function(id, data, dataformats, smart_selector)
{
    this.element        = $('#' + id);
    this.groups         = [];
    this.label          = this.element.find('label');
    this.dataformats    = dataformats;
    this.smart_selector = smart_selector;
    this.analyzer       = false;


    var panel = this;

    this.element.find('.add_button').click(function() {
        var group = {
            name: panel._generateGroupName(),
            inputs: {
            },
            outputs: {
            }
        };

        group.inputs[panel._generateInputName()] = {
            type: null,
        };

        panel._addGroup(group);
    });


    if (data.length > 0)
    {
        var unsynchronized_group = {
            name: null,
            inputs: {},
        };

        var group_index = 0;
        var all_input_names = [];

        while (group_index < data.length)
        {
            var group = data[group_index];

            // Ensure that the unsynchronized inputs are all in the last group
            if ((group.name === undefined) && (group_index > 0))
            {
                var input_names = Object.keys(group.inputs);

                for (var j = 0; j < input_names.length; ++j)
                    unsynchronized_group.inputs[input_names[j]] = group.inputs[input_names[j]];

                data.splice(group_index, 1);

                all_input_names = all_input_names.concat(input_names);
            }

            else
            {
                // Ensure that all groups have at least one input
                if (Object.keys(group.inputs).length == 0)
                {
                    var input_name = this._generateInputName(all_input_names);

                    group.inputs[input_name] = {
                        type: null,
                    };
                }

                // Ensure that the first group has at least one output
                if (group_index == 0)
                {
                    if (data[group_index].outputs === undefined)
                        data[group_index].outputs = {};

                    if (Object.keys(data[group_index].outputs).length == 0)
                    {
                        var output_name = this._generateOutputName(Object.keys(group.outputs));

                        group.outputs[output_name] = {
                            type: null,
                        };
                    }
                }

                all_input_names = all_input_names.concat(Object.keys(group.inputs));

                ++group_index;
            }
        }

        if (Object.keys(unsynchronized_group.inputs).length > 0)
            data.push(unsynchronized_group);


        // Add the group controls
        for (var i = 0; i < data.length; ++i)
            this._addGroup(data[i]);
    }
    else
    {
        this._addGroup({
            name: "main",
            inputs: {
                "input1": {
                    type: null,
                }
            },
            outputs: {
                "output1": {
                    type: null,
                }
            }
        });
    }
}


//----------------------------------------------------------------------------------------
// Change the display of the editor according to the type of algorithm (analyzer or not)
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype.setAnalyzer = function(analyzer)
{
    this.analyzer = analyzer;

    for (var i = 0; i < this.groups.length; ++i)
    {
        var group_entry = this.groups[i];

        if (analyzer)
        {
            $(group_entry.outputs.label).hide();
            $(group_entry.outputs.table).hide();
        }
        else if (i == 0)
        {
            $(group_entry.outputs.label).show();
            $(group_entry.outputs.table).show();
        }
    }
}


//----------------------------------------------------------------------------------------
// Returns a JSON representation of the inputs & outputs
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype.toJSON = function(errors_callback)
{
    var result = [];
    var errors = [];
    var all_group_names = [];
    var all_input_names = [];
    var all_output_names = [];

    for (var i = 0; i < this.groups.length; ++i)
    {
        var group_entry = this.groups[i];

        var dest_group = {};

        if (this.groups.length > 1)
        {
            if ((i < this.groups.length - 1) || !group_entry.unsynchronized.checked)
            {
                if (group_entry.name.value.length == 0)
                {
                    errors.push('No name specified for inputs group #' + (i + 1));
                    continue;
                }

                if (all_group_names.indexOf(group_entry.name.value) >= 0)
                {
                    errors.push('Several inputs groups have the same name: ' + group_entry.name.value);
                    continue;
                }

                all_group_names.push(group_entry.name.value);

                dest_group.name = group_entry.name.value;
            }
        }
        else
        {
            dest_group.name = 'main';
        }


        dest_group.inputs = {};

        for (var j = 0; j < group_entry.inputs.entries.length; ++j)
        {
            var input_entry = group_entry.inputs.entries[j];

            if (input_entry.name.value.length == 0)
            {
                errors.push('No name specified for the input #' + (j + 1) + ' of group "' + group_entry.name.value + '"');
                continue;
            }

            if (all_input_names.indexOf[input_entry.name.value] >= 0)
            {
                errors.push('Several inputs have the same name: ' + input_entry.name.value);
                continue;
            }

            all_input_names.push(input_entry.name.value);

            if (input_entry.dataformat === null)
            {
                errors.push('No data format specified for the input "' + input_entry.name.value + '"');
                continue;
            }

            dest_group.inputs[input_entry.name.value] = {
                type: input_entry.dataformat,
            };
        }


        if (!this.analyzer && (i == 0))
        {
            dest_group.outputs = {};

            for (var j = 0; j < group_entry.outputs.entries.length; ++j)
            {
                var output_entry = group_entry.outputs.entries[j];

                if (output_entry.name.value.length == 0)
                {
                    errors.push('No name specified for the output #' + (j + 1) + ' of group "' + group_entry.name.value + '"');
                    continue;
                }

                if (all_output_names.indexOf[output_entry.name.value] >= 0)
                {
                    errors.push('Several outputs have the same name: ' + output_entry.name.value);
                    continue;
                }

                all_output_names.push(output_entry.name.value);

                if (output_entry.dataformat === null)
                {
                    errors.push('No data format specified for the output "' + output_entry.name.value + '"');
                    continue;
                }

                dest_group.outputs[output_entry.name.value] = {
                    type: output_entry.dataformat,
                };
            }
        }

        result.push(dest_group);
    }

    if (!this.analyzer && (all_output_names.length == 0))
        errors.push('No output declared. An algorithm must have at least one output');

    if (errors.length > 0)
    {
        if (errors_callback)
            errors_callback(errors);

        return null;
    }

    return result;
}


//----------------------------------------------------------------------------------------
// Generate a new unique group name
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype._generateGroupName = function()
{
    var all_group_names = [];

    for (var i = 0; i < this.groups.length; ++i)
        all_group_names.push(this.groups[i].name.value);

    return this._generateUniqueName(all_group_names, 'group');
}


//----------------------------------------------------------------------------------------
// Generate a new unique input name
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype._generateInputName = function(all_input_names)
{
    if (all_input_names === undefined)
    {
        all_input_names = [];

        for (var i = 0; i < this.groups.length; ++i)
        {
            for (var j = 0; j < this.groups[i].inputs.entries.length; ++j)
                all_input_names.push(this.groups[i].inputs.entries[j].name.value);
        }
    }

    return this._generateUniqueName(all_input_names, 'input');
}


//----------------------------------------------------------------------------------------
// Generate a new unique output name
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype._generateOutputName = function(all_output_names)
{
    if (all_output_names === undefined)
    {
        all_output_names = [];

        for (var j = 0; j < this.groups[0].outputs.entries.length; ++j)
            all_output_names.push(this.groups[0].outputs.entries[j].name.value);
    }

    return this._generateUniqueName(all_output_names, 'output');
}


//----------------------------------------------------------------------------------------
// Generate a new unique output name
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype._generateUniqueName = function(all_names, prefix)
{
    var index = all_names.length + 1;
    var unique_name = prefix + index;

    while (all_names.indexOf(unique_name) >= 0)
    {
        ++index;
        unique_name = prefix + index;
    }

    return unique_name;
}


//----------------------------------------------------------------------------------------
// Add the controls related to a new group
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype._addGroup = function(data)
{
    function _createTable(container, target)
    {
        target.label = document.createElement('span');
        target.label.className = 'label table_label';
        container.appendChild(target.label);

        target.table = document.createElement('table');
        container.appendChild(target.table);

        var tbody = document.createElement('tbody');
        target.table.appendChild(tbody);

        target.last_row = document.createElement('tr');
        tbody.appendChild(target.last_row);

        var td = document.createElement('td');
        td.colSpan = 2;
        target.last_row.appendChild(td);

        target.add_button = document.createElement('span');
        target.add_button.className = 'add_button fa-stack';
        var background_icon = document.createElement('i');
        background_icon.className = 'fa fa-circle fa-stack-2x';
        target.add_button.appendChild(background_icon);
        var foreground_icon = document.createElement('i');
        foreground_icon.className = 'fa fa-plus fa-inverse fa-stack-1x';
        target.add_button.appendChild(foreground_icon);
        td.appendChild(target.add_button);
    }


    var panel = this;

    var group_entry = {
        container:            null,
        header:               null,
        name:                 null,
        unsynchronized:       null,
        unsynchronized_label: null,
        delete_button:        null,

        inputs: {
            label:      null,
            table:      null,
            last_row:   null,
            add_button: null,
            entries:    [],
        },

        outputs: {
            label:      null,
            table:      null,
            last_row:   null,
            add_button: null,
            entries:    [],
        },
    };

    group_entry.container = document.createElement('div');
    group_entry.container.className = 'col-sm-12 input_group';

    group_entry.header = document.createElement('p');
    group_entry.header.className = 'group';
    group_entry.container.appendChild(group_entry.header);

    var span = document.createElement('span');
    span.innerHTML = 'Group:';
    span.className = 'label';
    group_entry.header.appendChild(span);

    group_entry.name = document.createElement('input');
    group_entry.name.className = 'group';
    group_entry.header.appendChild(group_entry.name);

    if (data && data.name)
        group_entry.name.value = data.name;

    $(group_entry.name).change(name_check);


    group_entry.unsynchronized = document.createElement('input');
    group_entry.unsynchronized.className = 'unsynchronized';
    group_entry.unsynchronized.type = 'checkbox';
    group_entry.header.appendChild(group_entry.unsynchronized);

    group_entry.unsynchronized_label = document.createElement('span');
    group_entry.unsynchronized_label.innerHTML = 'Unsynchronized';
    group_entry.unsynchronized_label.className = 'unsynchronized';
    group_entry.header.appendChild(group_entry.unsynchronized_label);

    if (data)
        group_entry.unsynchronized.checked = (data.name === null);


    group_entry.delete_button = document.createElement('span');
    group_entry.delete_button.className = 'delete_button fa-stack';
    group_entry.delete_button.title = 'Delete this group';
    var background_icon = document.createElement('i');
    background_icon.className = 'fa fa-circle fa-stack-2x';
    group_entry.delete_button.appendChild(background_icon);
    var foreground_icon = document.createElement('i');
    foreground_icon.className = 'fa fa-minus fa-inverse fa-stack-1x';
    group_entry.delete_button.appendChild(foreground_icon);
    group_entry.header.appendChild(group_entry.delete_button);


    $(group_entry.unsynchronized).change(function() {
        if (group_entry.unsynchronized.checked)
            $(group_entry.name).hide();
        else
            $(group_entry.name).show();
    });

    $(group_entry.delete_button).mouseover(function() {
        $(group_entry.container).addClass('deletion_information');
    });

    $(group_entry.delete_button).mouseout(function() {
        $(group_entry.container).removeClass('deletion_information');
    });

    $(group_entry.delete_button).click(function() {
        for (var i = 0; i < panel.groups.length; ++i)
        {
            if (panel.groups[i].container == group_entry.container)
            {
                panel.element[0].removeChild(group_entry.container);
                panel.groups.splice(i, 1);

                if (panel.groups[0].outputs.entries.length == 0)
                {
                    panel._addConnector(panel.groups[0].outputs, {
                        name: panel._generateOutputName(),
                        type: null
                    });
                }

                panel._update();
                break;
            }
        }
    });


    _createTable(group_entry.container, group_entry.inputs);
    group_entry.inputs.label.innerHTML = 'Inputs:';
    group_entry.inputs.table.className = 'inputs';
    group_entry.inputs.add_button.title = 'Add a new input';

    _createTable(group_entry.container, group_entry.outputs);
    group_entry.outputs.label.innerHTML = 'Outputs:';
    group_entry.outputs.label.className += ' outputs';
    group_entry.outputs.table.className = 'outputs';
    group_entry.outputs.add_button.title = 'Add a new output';

    if (this.analyzer || (this.groups.length > 0))
    {
        $(group_entry.outputs.label).hide();
        $(group_entry.outputs.table).hide();
    }


    $(group_entry.inputs.add_button).click(function() {
        panel._addConnector(group_entry.inputs, {
            name: panel._generateInputName(),
            type: null,
        });
        panel._update();
    });


    $(group_entry.outputs.add_button).click(function() {
        panel._addConnector(group_entry.outputs, {
            name: panel._generateOutputName(),
            type: null,
        });
        panel._update();
    });


    if (this.groups.length == 0)
    {
        $(group_entry.container).insertAfter(this.label);
        this.groups.push(group_entry);
    }
    else if (this.groups[this.groups.length - 1].unsynchronized.checked)
    {
        $(group_entry.container).insertBefore( $(this.groups[this.groups.length - 1].container) );
        this.groups.splice(this.groups.length - 1, 0, group_entry);
    }
    else
    {
        $(group_entry.container).insertAfter( $(this.groups[this.groups.length - 1].container) );
        this.groups.push(group_entry);
    }


    var input_names = Object.keys(data.inputs);
    for (var i = 0; i < input_names.length; ++i)
    {
        var input_data = data.inputs[input_names[i]];
        input_data.name = input_names[i];
        this._addConnector(group_entry.inputs, input_data);
    }

    if (data.outputs)
    {
        var output_names = Object.keys(data.outputs);
        for (var i = 0; i < output_names.length; ++i)
        {
            var output_data = data.outputs[output_names[i]];
            output_data.name = output_names[i];
            this._addConnector(group_entry.outputs, output_data);
        }
    }

    this._update();
}


//----------------------------------------------------------------------------------------
// Add the controls related to a new input or output
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype._addConnector = function(target, data)
{
    var entry = {
        row:                null,
        name:               null,
        dataformat_element: null,
        dataformat:         null,
        delete_button:      null,
    };

    entry.row = document.createElement('tr');
    $(entry.row).insertBefore(target.last_row);

    var td = document.createElement('td');
    entry.row.appendChild(td);

    entry.name = document.createElement('input');
    td.appendChild(entry.name);

    if (data.name)
        entry.name.value = data.name;

    td = document.createElement('td');
    entry.row.appendChild(td);

    entry.dataformat_element = document.createElement('span');
    entry.dataformat_element.className = 'dataformat';
    td.appendChild(entry.dataformat_element);

    if (data.type)
    {
        $(entry.dataformat_element).text(data.type);
        entry.dataformat = data.type;
    }
    else
    {
        $(entry.dataformat_element).text('<select one data format>');
        $(entry.dataformat_element).addClass('unselected');
    }

    td = document.createElement('td');
    entry.row.appendChild(td);

    entry.delete_button = document.createElement('span');
    entry.delete_button.className = 'delete_button fa-stack';
    entry.delete_button.title = 'Delete this entry';
    var background_icon = document.createElement('i');
    background_icon.className = 'fa fa-circle fa-stack-2x';
    entry.delete_button.appendChild(background_icon);
    var foreground_icon = document.createElement('i');
    foreground_icon.className = 'fa fa-minus fa-inverse fa-stack-1x';
    entry.delete_button.appendChild(foreground_icon);
    td.appendChild(entry.delete_button);

    target.entries.push(entry);


    var panel = this;

    $(entry.name).change(name_check);


    $(entry.dataformat_element).click(function() {
        panel.smart_selector.onEntrySelected = function(dataformat_name) {
            $(panel.current_connector.dataformat_element).text(dataformat_name);
            $(panel.current_connector.dataformat_element).removeClass('unselected');

            panel.current_connector.dataformat = dataformat_name;

            panel.current_connector = null;
        };


        panel.current_connector = entry;

        var entries = [];
        for (var i = 0; i < panel.dataformats.length; ++i)
        {
            entries.push({
                name:        panel.dataformats[i].name,
                description: panel.dataformats[i].short_description,
            });
        }

        var left = $(target.table).offset().left + $(target.table).width() + 10;
        var top = $(entry.row).offset().top;

        panel.smart_selector.display(entries, left, top);
    });


    $(entry.delete_button).click(function() {
        for (var i = 0; i < target.entries.length; ++i)
        {
            if (target.entries[i].row == entry.row)
            {
                $(target.table).find('tbody')[0].removeChild(entry.row);
                target.entries.splice(i, 1);
                panel._update();
                break;
            }
        }
    });
}


//----------------------------------------------------------------------------------------
// Ensure that all elements of the the inputs editor are correctly displayed
//----------------------------------------------------------------------------------------
beat.contributions.editor.InputsEditor.prototype._update = function()
{
    if (this.groups.length == 1)
        $(this.groups[0].header).hide();
    else
        $(this.groups[0].header).show();

    for (var i = 0; i < this.groups.length; ++i)
    {
        var group_entry = this.groups[i];

        if (group_entry.inputs.entries.length == 1)
            $(group_entry.inputs.entries[0].delete_button).hide();
        else
            $(group_entry.inputs.entries[0].delete_button).show();

        if (group_entry.outputs.entries.length == 1)
            $(group_entry.outputs.entries[0].delete_button).hide();
        else if (group_entry.outputs.entries.length > 1)
            $(group_entry.outputs.entries[0].delete_button).show();

        if (i < this.groups.length - 1)
        {
            group_entry.unsynchronized.checked = false;
            $(group_entry.unsynchronized).hide();
            $(group_entry.unsynchronized_label).hide();
        }
        else
        {
            $(group_entry.unsynchronized).show();
            $(group_entry.unsynchronized_label).show();

            if (group_entry.unsynchronized.checked)
                $(group_entry.name).hide();
            else
                $(group_entry.name).show();
        }
    }

    if (!this.analyzer) {
      $(this.groups[0].outputs.label).show();
      $(this.groups[0].outputs.table).show();
    }
}



/******************************* CLASS: ResultsEditor ***********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.contributions.editor.ResultsEditor = function(id, results, chart_types)
{
    this.element     = $('#' + id);
    this.entries     = [];
    this.table       = this.element.find('table');
    this.chart_types = chart_types;


    var panel = this;
    this.element.find('.add_button').click(function() {
        panel._addResult();
    });


    var result_names = Object.keys(results);
    for (var i = 0; i < result_names.length; ++i)
    {
        var result = results[result_names[i]];
        result.name = result_names[i];
        this._addResult(result);
    }
}


//----------------------------------------------------------------------------------------
// Returns a JSON representation of the parameters
//----------------------------------------------------------------------------------------
beat.contributions.editor.ResultsEditor.prototype.getResults = function(error_callback)
{
    if (this.entries.length == 0)
    {
        if (error_callback)
            error_callback(['No result field declared']);
        return null;
    }

    var results = {};

    for (var i = 0; i < this.entries.length; ++i)
    {
        var entry = this.entries[i];

        var name = entry.name.value.trim().replace(/\s/g, '_');

        if (name.replace(/_/g, '').length == 0)
        {
            if (error_callback)
                error_callback(['Invalid name for the result field #' + i]);
            return null;
        }

        results[name] = {
            type:    entry.type.value,
            display: entry.primary.checked,
        };
    }

    return results;
}


//----------------------------------------------------------------------------------------
// Display the results editor
//----------------------------------------------------------------------------------------
beat.contributions.editor.ResultsEditor.prototype.show = function()
{
    return this.element.show();
}


//----------------------------------------------------------------------------------------
// Hide the results editor
//----------------------------------------------------------------------------------------
beat.contributions.editor.ResultsEditor.prototype.hide = function()
{
    return this.element.hide();
}


//----------------------------------------------------------------------------------------
// Add the controls related to a new result field
//----------------------------------------------------------------------------------------
beat.contributions.editor.ResultsEditor.prototype._addResult = function(data)
{
    var panel = this;

    var entry = {
        row: null,
        name: null,
        type: null,
        primary: null,
        delete_button: null,
    };

    entry.row = document.createElement('tr');


    var cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.name = document.createElement('input');
    entry.name.className = 'name';
    cell.appendChild(entry.name);

    if (data)
        entry.name.value = data.name;

    $(entry.name).change(name_check);


    cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.type = document.createElement('select');
    cell.appendChild(entry.type);

    var valid_types = ['float32', 'int32', 'bool', 'string'].concat(this.chart_types.map(function(x) { return x.name; }));

    for (var  i = 0; i < valid_types.length; ++i)
    {
        var option = document.createElement('option');
        option.text = valid_types[i];
        option.value = valid_types[i];

        if ((data && (valid_types[i] == data.type)) ||
            (!data && (valid_types[i] == 'float32')))
        {
            option.selected = true;
        }

        entry.type.appendChild(option);
    }


    cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.primary = document.createElement('input');
    entry.primary.type = 'checkbox';
    cell.appendChild(entry.primary);

    if (data && (data.display !== undefined))
        entry.primary.checked = data.display;


    cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.delete_button = document.createElement('span');
    entry.delete_button.className = 'delete_button fa-stack';
    entry.delete_button.title = 'Delete this entry';
    var background_icon = document.createElement('i');
    background_icon.className = 'fa fa-circle fa-stack-2x';
    entry.delete_button.appendChild(background_icon);
    var foreground_icon = document.createElement('i');
    foreground_icon.className = 'fa fa-minus fa-inverse fa-stack-1x';
    entry.delete_button.appendChild(foreground_icon);
    entry.delete_button.title = 'Delete this result field';
    cell.appendChild(entry.delete_button);

    $(entry.delete_button).click(function() { panel._deleteResult(entry); });


    var tbody = this.table.find('tbody')[0];

    tbody.appendChild(entry.row);
    this.entries.push(entry);

    this.table.show();
}


//----------------------------------------------------------------------------------------
// Delete all the controls related to a result field
//----------------------------------------------------------------------------------------
beat.contributions.editor.ResultsEditor.prototype._deleteResult = function(entry)
{
    this.table.find('tbody')[0].removeChild(entry.row);

    for (var i = 0; i < this.entries.length; ++i)
    {
        if (this.entries[i] == entry)
        {
            this.entries.splice(i, 1);
            break;
        }
    }

    if (this.entries.length == 0)
        this.table.hide();
}



/****************************** CLASS: ParametersEditor *********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor = function(id, parameters, callbacks)
{
    this.element   = $('#' + id);
    this.entries   = [];
    this.table     = this.element.find('table');
    this.callbacks = callbacks;


    var panel = this;
    this.element.find('.add_button').click(function() {
        panel._addParameter();

        if (panel.callbacks.addMethodToSourceCode)
            panel.callbacks.addMethodToSourceCode();
    });


    var parameter_names = Object.keys(parameters);
    for (var i = 0; i < parameter_names.length; ++i)
    {
        var parameter = parameters[parameter_names[i]];
        parameter.name = parameter_names[i];
        this._addParameter(parameter);
    }
}


//----------------------------------------------------------------------------------------
// Returns a JSON representation of the parameters
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype.getParameters = function()
{
    var parameters = {};

    for (var i = 0; i < this.entries.length; ++i)
    {
        var entry = this.entries[i];

        var name = entry.name.value.trim().replace(/\s/g, '_');

        if (name.replace(/_/g, '').length == 0)
        {
            if (this.callbacks.displayErrors)
                this.callbacks.displayErrors(['Invalid name for the parameter #' + i]);
            return null;
        }

        var json_parameter = {
            type:        entry.type.value,
            description: entry.description.value,
        };

        if (entry.type.value == 'bool')
        {
            json_parameter.default = entry.default_bool_value.checked;
        }
        else if ((entry.type.value == 'string') || entry.choices.checked)
        {
            json_parameter.default = entry.default_choice_value.value;
            json_parameter.choice = [];

            var choices = entry.choices_list.value.split('\n');

            for (var j = 0; j < choices.length; ++j)
            {
                if (choices[j].trim().length > 0)
                    json_parameter.choice.push(choices[j].trim());
            }
        }
        else if (entry.range.checked)
        {
            var conversion = null;
            var limit_minimum = null;
            var limit_maximum = null;

            if (entry.type.value == 'uint8')
            {
                conversion = parseInt;
                limit_minimum = 0;
                limit_maximum = 255;
            }
            else if (entry.type.value == 'uint16')
            {
                conversion = parseInt;
                limit_minimum = 0;
                limit_maximum = Math.pow(2, 16) - 1;
            }
            else if (entry.type.value == 'uint32')
            {
                conversion = parseInt;
                limit_minimum = 0;
                limit_maximum = Math.pow(2, 32) - 1;
            }
            else if (entry.type.value == 'uint64')
            {
                conversion = parseInt;
                limit_minimum = 0;
                limit_maximum = Math.pow(2, 64) - 1;
            }
            else if (entry.type.value == 'int8')
            {
                conversion = parseInt;
                limit_minimum = 127;
                limit_maximum = -128;
            }
            else if (entry.type.value == 'int16')
            {
                conversion = parseInt;
                limit_minimum = Math.pow(2, 15 - 1);
                limit_maximum = -Math.pow(2, 15);
            }
            else if (entry.type.value == 'int32')
            {
                conversion = parseInt;
                limit_minimum = Math.pow(2, 31 - 1);
                limit_maximum = -Math.pow(2, 31);
            }
            else if (entry.type.value == 'int64')
            {
                conversion = parseInt;
                limit_minimum = Math.pow(2, 63 - 1);
                limit_maximum = -Math.pow(2, 63);
            }
            else
            {
                conversion = parseFloat;
                limit_minimum = -Infinity;
                limit_maximum = Infinity;
            }

            var minimum = conversion(entry.minimum.value.trim());
            var maximum = conversion(entry.maximum.value.trim());
            var default_value = conversion(entry.default_value.value.trim());

            if (isNaN(minimum) || isNaN(maximum) ||
                (minimum < limit_minimum) || (maximum > limit_maximum))
            {
                if (this.callbacks.displayErrors)
                    this.callbacks.displayErrors(["Invalid range for the parameter '" + name + "'"]);
                return null;
            }

            if (isNaN(default_value) ||
                (default_value < limit_minimum) || (default_value > limit_maximum))
            {
                if (this.callbacks.displayErrors)
                    this.callbacks.displayErrors(["Invalid default value for the parameter '" + name + "'"]);
                return null;
            }

            json_parameter.default = default_value;
            json_parameter.range = [minimum, maximum];
        }
        else
        {
            if ((entry.type.value == 'float32') || (entry.type.value == 'float64'))
                var default_value = parseFloat(entry.default_value.value.trim());
            else
                var default_value = parseInt(entry.default_value.value.trim());

            if (isNaN(default_value))
            {
                if (this.callbacks.displayErrors)
                    this.callbacks.displayErrors(["Invalid default value for the parameter '" + name + "'"]);
                return null;
            }

            json_parameter.default = default_value;
        }

        parameters[name] = json_parameter;
    }

    return parameters;
}


//----------------------------------------------------------------------------------------
// Called when the type of a parameter is changed
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._typeChanged = function(entry)
{
    if (entry.type.value == 'bool')
    {
        $(entry.range).hide();
        $(entry.minimum).hide();
        $(entry.range_separator).hide();
        $(entry.maximum).hide();
        $(entry.choices).hide();
        $(entry.choices_list).hide();
        $(entry.default_value).hide();
        $(entry.default_bool_value).show();
        $(entry.default_choice_value).hide();
    }
    else if (entry.type.value == 'string')
    {
        $(entry.range).hide();
        $(entry.minimum).hide();
        $(entry.range_separator).hide();
        $(entry.maximum).hide();
        $(entry.choices).hide();
        $(entry.choices_list).show();
        $(entry.default_value).hide();
        $(entry.default_bool_value).hide();
        $(entry.default_choice_value).show();
    }
    else
    {
        $(entry.range).show();
        $(entry.choices).show();

        this._rangeClicked(entry);
        this._choicesClicked(entry);

        $(entry.default_value).show();
        $(entry.default_bool_value).hide();
        $(entry.default_choice_value).hide();
    }

    this._updateTable();
}


//----------------------------------------------------------------------------------------
// Called when the 'range' checkbox of a parameter is clicked
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._rangeClicked = function(entry)
{
    if (entry.range.checked)
    {
        $(entry.minimum).show();
        $(entry.range_separator).show();
        $(entry.maximum).show();

        $(entry.choices).hide();
        $(entry.choices_list).hide();
    }
    else
    {
        $(entry.minimum).hide();
        $(entry.range_separator).hide();
        $(entry.maximum).hide();

        $(entry.choices).show();
        $(entry.choices_list).hide();
    }

    this._updateChoicesColumn();
}


//----------------------------------------------------------------------------------------
// Called when the 'choices' checkbox of a parameter is clicked
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._choicesClicked = function(entry)
{
    if (entry.choices.checked)
    {
        $(entry.choices_list).show();

        $(entry.range).hide();
        $(entry.minimum).hide();
        $(entry.range_separator).hide();
        $(entry.maximum).hide();

        $(entry.default_choice_value).show();
        $(entry.default_value).hide();
        $(entry.default_bool_value).hide();
    }
    else
    {
        $(entry.choices_list).hide();

        $(entry.range).show();
        $(entry.minimum).hide();
        $(entry.range_separator).hide();
        $(entry.maximum).hide();

        $(entry.default_choice_value).hide();
        $(entry.default_value).show();
        $(entry.default_bool_value).hide();
    }

    this._updateRangeColumn();
}


//----------------------------------------------------------------------------------------
// Called when the possible choices for a parameter are changed
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._updateChoices = function(entry)
{
    var lines = entry.choices_list.value.split('\n');

    var index = 0;
    for (var i = 0; i < lines.length; ++i)
    {
        var line = lines[i].trim();
        if (line.length == 0)
            continue;

        if (index >= entry.default_choice_value.options.length)
        {
            var option = document.createElement('option');
            option.text = line;
            option.value = line;
            entry.default_choice_value.appendChild(option);
        }
        else
        {
            entry.default_choice_value.options[index].text = line;
            entry.default_choice_value.options[index].value = line;
        }

        ++index;
    }

    for (var i = index; i < entry.default_choice_value.options.length; ++i)
        entry.default_choice_value.removeChild(entry.default_choice_value.options[i]);
}


//----------------------------------------------------------------------------------------
// Add the controls related to a new parameter
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._addParameter = function(data)
{
    var panel = this;

    var entry = {
        row: null,
        name: null,
        type: null,
        range: null,
        minimum: null,
        range_separator: null,
        maximum: null,
        choices: null,
        choices_list: null,
        default_value: null,
        default_bool_value: null,
        default_choice_value: null,
        description: null,
        delete_button: null,
    };

    entry.row = document.createElement('tr');


    var cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.name = document.createElement('input');
    entry.name.className = 'name';
    cell.appendChild(entry.name);

    if (data)
        entry.name.value = data.name;

    $(entry.name).change(name_check);


    cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.type = document.createElement('select');
    cell.appendChild(entry.type);

    var valid_types = ['uint8', 'uint16', 'uint32', 'uint64',
                       'int8', 'int16', 'int32', 'int64',
                       'float32', 'float64', 'bool', 'string'
                      ];

    for (var  i = 0; i < valid_types.length; ++i)
    {
        var option = document.createElement('option');
        option.text = valid_types[i];
        option.value = valid_types[i];

        if ((data && (valid_types[i] == data.type)) ||
            (!data && (valid_types[i] == 'uint32')))
        {
            option.selected = true;
        }

        entry.type.appendChild(option);
    }

    $(entry.type).change(function() { panel._typeChanged(entry); });


    cell = document.createElement('td');
    cell.className = 'range_column';
    entry.row.appendChild(cell);

    entry.range = document.createElement('input');
    entry.range.type = 'checkbox';
    cell.appendChild(entry.range);

    entry.minimum = document.createElement('input');
    entry.minimum.className = 'range';
    $(entry.minimum).hide();
    cell.appendChild(entry.minimum);

    entry.range_separator = document.createElement('span');
    $(entry.range_separator).text('-');
    $(entry.range_separator).hide();
    cell.appendChild(entry.range_separator);

    entry.maximum = document.createElement('input');
    entry.maximum.className = 'range';
    $(entry.maximum).hide();
    cell.appendChild(entry.maximum);

    $(entry.range).change(function() { panel._rangeClicked(entry); });


    cell = document.createElement('td');
    cell.className = 'choices_column';
    entry.row.appendChild(cell);

    entry.choices = document.createElement('input');
    entry.choices.type = 'checkbox';
    cell.appendChild(entry.choices);

    cell = document.createElement('td');
    cell.className = 'choices_column second';
    entry.row.appendChild(cell);

    entry.choices_list = document.createElement('textarea');
    entry.choices_list.className = 'choices';
    $(entry.choices_list).hide();
    cell.appendChild(entry.choices_list);

    entry.choices_list.value = 'value1\nvalue2\nvalue3';

    $(entry.choices).change(function() { panel._choicesClicked(entry); });

    $(entry.choices_list).keyup(function() {
        panel._updateChoices(entry);
    });


    cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.default_value = document.createElement('input');
    entry.default_value.className = 'default';
    cell.appendChild(entry.default_value);

    entry.default_bool_value = document.createElement('input');
    entry.default_bool_value.type = 'checkbox';
    $(entry.default_bool_value).hide();
    cell.appendChild(entry.default_bool_value);

    entry.default_choice_value = document.createElement('select');
    $(entry.default_choice_value).hide();
    cell.appendChild(entry.default_choice_value);


    cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.description = document.createElement('input');
    entry.description.className = 'description';
    cell.appendChild(entry.description);

    if (data && (data.description !== undefined))
        entry.description.value = data.description;


    cell = document.createElement('td');
    entry.row.appendChild(cell);

    entry.delete_button = document.createElement('span');
    entry.delete_button.className = 'delete_button fa-stack';
    entry.delete_button.title = 'Delete this entry';
    var background_icon = document.createElement('i');
    background_icon.className = 'fa fa-circle fa-stack-2x';
    entry.delete_button.appendChild(background_icon);
    var foreground_icon = document.createElement('i');
    foreground_icon.className = 'fa fa-minus fa-inverse fa-stack-1x';
    entry.delete_button.appendChild(foreground_icon);
    entry.delete_button.title = 'Delete this parameter';
    cell.appendChild(entry.delete_button);

    $(entry.delete_button).click(function() { panel._deleteParameter(entry); });


    var tbody = this.table.find('tbody')[0];

    tbody.appendChild(entry.row);
    this.entries.push(entry);


    if (data)
    {
        this._typeChanged(entry);

        if (data.range !== undefined)
        {
            entry.range.checked = true;
            entry.minimum.value = '' + data.range[0];
            entry.maximum.value = '' + data.range[1];
            entry.default_value.value = '' + data.default;
            this._rangeClicked(entry);
        }
        else if (data.choice !== undefined)
        {
            entry.choices.checked = true;

            var choices = '';
            var selected_index = 0;
            for (var i = 0; i < data.choice.length; ++i)
            {
                choices += data.choice[i];
                if (i < data.choice.length - 1)
                    choices += '\n';

                if (data.choice[i] == data.default)
                    selected_index = i;
            }
            entry.choices_list.value = choices;

            this._choicesClicked(entry);
            this._updateChoices(entry);

            entry.default_choice_value.selectedIndex = selected_index;
        }
        else if (data.type == 'bool')
        {
            entry.default_bool_value.checked = data.default;
        }
        else
        {
            entry.default_value.value = data.default;
        }
    }
    else
    {
        this._updateChoices(entry);
    }

    this._updateTable();

    this.table.show();
}


//----------------------------------------------------------------------------------------
// Delete all the controls related to a parameter
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._deleteParameter = function(entry)
{
    this.table.find('tbody')[0].removeChild(entry.row);

    for (var i = 0; i < this.entries.length; ++i)
    {
        if (this.entries[i] == entry)
        {
            this.entries.splice(i, 1);
            break;
        }
    }

    this._updateTable();
}


//----------------------------------------------------------------------------------------
// Display/hide the 'range' column if necessary
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._updateRangeColumn = function()
{
    var must_hide = true;
    for (var i = 0; i < this.entries.length; ++i)
    {
        if ((this.entries[i].type.value != 'bool') && !this.entries[i].choices.checked)
        {
            must_hide = false;
            break;
        }
    }

    if (must_hide)
        this.table.find('.range_column').hide();
    else
        this.table.find('.range_column').show();
}


//----------------------------------------------------------------------------------------
// Display/hide the 'choices' column if necessary
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._updateChoicesColumn = function()
{
    var must_hide = true;
    for (var i = 0; i < this.entries.length; ++i)
    {
        if ((this.entries[i].type.value != 'bool') && !this.entries[i].range.checked)
        {
            must_hide = false;
            break;
        }
    }

    if (must_hide)
        this.table.find('.choices_column').hide();
    else
        this.table.find('.choices_column').show();
}


//----------------------------------------------------------------------------------------
// Ensure that the status of the table is up-to-date
//----------------------------------------------------------------------------------------
beat.contributions.editor.ParametersEditor.prototype._updateTable = function()
{
    this._updateRangeColumn();
    this._updateChoicesColumn();

    if (this.entries.length == 0)
    {
        this.table.hide();

        if (this.callbacks.removeMethodFromSourceCode)
            this.callbacks.removeMethodFromSourceCode();
    }
}



/****************************** CLASS: LibrariesEditor *********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.contributions.editor.LibrariesEditor = function(id, data, libraries, smart_selector)
{
    this.element        = $('#' + id);
    this.entries        = [];
    this.table          = this.element.find('table');
    this.libraries      = libraries;
    this.smart_selector = smart_selector;


    var panel = this;

    this.element.find('.add_button').click(function() {
        panel._addLibrary();
    });


    var library_names = Object.keys(data);
    for (var i = 0; i < library_names.length; ++i)
        this._addLibrary(library_names[i], data[library_names[i]]);
}


//----------------------------------------------------------------------------------------
// Returns a JSON representation of the libraries
//----------------------------------------------------------------------------------------
beat.contributions.editor.LibrariesEditor.prototype.getLibraries = function(error_callback)
{
    var libraries = {};

    for (var i = 0; i < this.entries.length; ++i)
    {
        var entry = this.entries[i];

        var name = entry.name.value.trim().replace(/\s/g, '_');

        if (name.replace(/_/g, '').length == 0)
        {
            if (error_callback)
                error_callback(['Invalid name for the library #' + i]);
            return null;
        }

        if (entry.library === null)
        {
            if (error_callback)
                error_callback(['No library selected for "' + name + '"']);
            return null;
        }

        libraries[name] = entry.library;
    }

    return libraries;
}


//----------------------------------------------------------------------------------------
// Add the controls related to a new input or output
//----------------------------------------------------------------------------------------
beat.contributions.editor.LibrariesEditor.prototype._addLibrary = function(name, library)
{
    var entry = {
        row:             null,
        name:            null,
        library_element: null,
        library:         null,
        delete_button:   null,
    };

    entry.row = document.createElement('tr');
    this.table.find('tbody')[0].appendChild(entry.row);

    var td = document.createElement('td');
    entry.row.appendChild(td);

    entry.name = document.createElement('input');
    td.appendChild(entry.name);

    if (name)
        entry.name.value = name;

    td = document.createElement('td');
    entry.row.appendChild(td);

    entry.library_element = document.createElement('span');
    entry.library_element.className = 'library';
    td.appendChild(entry.library_element);

    if (library)
    {
        $(entry.library_element).text(library);
        entry.library = library;
    }
    else
    {
        $(entry.library_element).text('<select one library>');
        $(entry.library_element).addClass('unselected');
    }

    td = document.createElement('td');
    entry.row.appendChild(td);

    entry.delete_button = document.createElement('span');
    entry.delete_button.className = 'delete_button fa-stack';
    entry.delete_button.title = 'Delete this entry';
    var background_icon = document.createElement('i');
    background_icon.className = 'fa fa-circle fa-stack-2x';
    entry.delete_button.appendChild(background_icon);
    var foreground_icon = document.createElement('i');
    foreground_icon.className = 'fa fa-minus fa-inverse fa-stack-1x';
    entry.delete_button.appendChild(foreground_icon);
    td.appendChild(entry.delete_button);

    this.entries.push(entry);


    var panel = this;

    $(entry.name).change(name_check);


    $(entry.library_element).click(function() {
        panel.smart_selector.onEntrySelected = function(library_name) {
            $(panel.current_entry.library_element).text(library_name);
            $(panel.current_entry.library_element).removeClass('unselected');

            panel.current_entry.library = library_name;

            panel.current_entry = null;
        };

        panel.current_entry = entry;

        var entries = [];
        for (var i = 0; i < panel.libraries.length; ++i)
        {
            entries.push({
                name:        panel.libraries[i].name,
                description: panel.libraries[i].short_description,
            });
        }

        var left = panel.table.offset().left + panel.table.width() + 10;
        var top = $(entry.row).offset().top;

        panel.smart_selector.display(entries, left, top);
    });


    $(entry.delete_button).click(function() {
        for (var i = 0; i < panel.entries.length; ++i)
        {
            if (panel.entries[i].row == entry.row)
            {
                panel.table.find('tbody')[0].removeChild(entry.row);
                panel.entries.splice(i, 1);
                panel._update();
                break;
            }
        }
    });

    panel._update();
}


//----------------------------------------------------------------------------------------
// Ensure that all elements of the the libraries editor are correctly displayed
//----------------------------------------------------------------------------------------
beat.contributions.editor.LibrariesEditor.prototype._update = function()
{
    if (this.entries.length == 0)
        this.table.hide();
    else
        this.table.show();
}
