#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import os
import shutil

import simplejson as json

from django.contrib.auth.models import User
from django.conf import settings

from ...dataformats.models import DataFormat
from ...common.testutils import BaseTestCase, tearDownModule
from ...team.models import Team

from ..models import Algorithm

from beat.core.algorithm import Storage

class AlgorithmsBaseTestCase(BaseTestCase):
    class Meta:
        model = Algorithm

    def setUp(self):
        if os.path.exists(settings.ALGORITHMS_ROOT):
            shutil.rmtree(settings.ALGORITHMS_ROOT)


    def tearDown(self):
        if os.path.exists(settings.ALGORITHMS_ROOT):
            shutil.rmtree(settings.ALGORITHMS_ROOT)


class AlgorithmsCreationFunction(AlgorithmsBaseTestCase):

    def test_without_code(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user,
            name='algorithm1',
            short_description='some description',
            )

        assert algorithm, errors

        self.assertEqual(algorithm.endpoints.count(), 0)

        storage = Storage(settings.PREFIX,
                algorithm.fullname())
        storage.language = 'python'

        assert storage.exists()

        self.assertEqual(algorithm.author, user)

        declaration = storage.json.load()
        self.assertNotEqual(len(declaration), 0)

        declaration = json.loads(declaration)
        storage.language = declaration['language']

        code = storage.code.try_load()
        self.assertNotEqual(len(code), 0)


    def test_with_code(self):
        declaration = """{
  "description": "some description",
  "language": "python",
  "splittable": false,
  "groups": [
    {
      "name": "channel1",
      "inputs": {
        "a": { "type": "johndoe/single_integer/1" },
        "b": { "type": "johndoe/single_integer/1" }
      },
      "outputs": {
        "sum": { "type": "johndoe/single_integer/1" }
      }
    }
  ],
  "parameters": {
    "dummy": {
      "type": "int32",
      "default": "10"
    }
  }
}"""

        code = """class Algorithm:

    def process(self, inputs, outputs):
        return True

"""

        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')

        (dataformat1, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='single_integer',
            short_description='short description 1',
            )

        assert dataformat1, errors

        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user,
            name='algorithm1',
            declaration=declaration,
            code=code,
            )

        assert algorithm, errors

        storage = Storage(settings.PREFIX,
                algorithm.fullname())
        storage.language = 'python'

        assert storage.exists()

        self.assertEqual(algorithm.author, user)
        self.assertTrue(algorithm.parameters)
        self.assertTrue(len(algorithm.parameters) > 0)

        parameters = json.loads(algorithm.parameters)
        self.assertEqual(len(parameters), 1)
        self.assertEqual(parameters[0]['name'], 'dummy')

        self.assertEqual(algorithm.endpoints.count(), 3)

        endpoint = algorithm.endpoints.filter(name='a')[0]
        self.assertTrue(endpoint.input)
        self.assertEqual(endpoint.dataformat, dataformat1)
        self.assertEqual(endpoint.channel, 'channel1')

        endpoint = algorithm.endpoints.filter(name='b')[0]
        self.assertTrue(endpoint.input)
        self.assertEqual(endpoint.dataformat, dataformat1)
        self.assertEqual(endpoint.channel, 'channel1')

        endpoint = algorithm.endpoints.filter(name='sum')[0]
        self.assertFalse(endpoint.input)
        self.assertEqual(endpoint.dataformat, dataformat1)
        self.assertEqual(endpoint.channel, 'channel1')

        in_storage = json.loads(storage.json.load())
        expected = json.loads(declaration)

        assert in_storage == expected, "There are differences on the " \
                "declaration (beyond white spaces): %r != %r" %  \
                (in_storage, expected)

        # set storage language so it can find the code
        storage.language = in_storage['language']

        in_storage = storage.code.try_load()
        expected = code

        assert in_storage == expected, "There are differences on the " \
                "code code: %r != %r" % (in_storage, expected)


class AlgorithmsTestCase(AlgorithmsBaseTestCase):
    def setup_algorithms(self, declaration, code):
        user1 = User.objects.get(username='jackdoe')
        user2 = User.objects.get(username='johndoe')

        team1 = Team.objects.get(name="teamdoe", owner=user1)
        team2 = Team.objects.get(name="teamdoe2", owner=user1)

        # Personal
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='personal',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        # Usable by one user
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='usable_by_one_user',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        algorithm.share(public=False, users=[user2])

        # Usable by one team
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='usable_by_one_team',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        algorithm.share(public=False, teams=[team1])

        # Usable by teams
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='usable_by_teams',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        algorithm.share(public=False, teams=[team1, team2])

        # Usable by all
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='usable_by_all',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        algorithm.share(public=False)

        # Accessible to one user
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='public_for_one_user',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        # algorithm.shared_with.add(user2)
        algorithm.share(public=True, users=[user2])

        # Accessible to one team
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='public_for_one_team',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        # algorithm.shared_with_team.add(team1)
        algorithm.share(public=True, teams=[team1])

        # Accessible to teams
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='public_for_teams',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        algorithm.share(public=True, teams=[team1, team2])

        # Accessible to all
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user1,
            name='public_for_all',
            short_description='',
            declaration=declaration,
            code=code,
            )
        assert algorithm, errors

        algorithm.share(public=True)


class AlgorithmsAccessibilityFunctionsBase(AlgorithmsTestCase):

    def setUp(self):
        super(AlgorithmsAccessibilityFunctionsBase, self).setUp()

        declaration = """{
  "language": "python",
  "splittable": false,
  "groups": [
    {
      "inputs": {
        "in": { "type": "jackdoe/public_for_all/1" }
      },
      "outputs": {
        "out": { "type": "jackdoe/public_for_all/1" }
      }
    }
  ],
  "parameters": {
  }
}"""
        code = """class Algorithm:

    def process(self, inputs, outputs):
        return True

"""
        # Users
        user1 = User.objects.create_user('jackdoe', 'jackdoe@test.org', '1234')
        user2 = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        user3 = User.objects.create_user('janedoe', 'janedoe@test.org', '1234')
        User.objects.create_user('jacquelinedoe', 'jacquelinedoe@test.org', '1234')

        team1 = Team.objects.create(name="teamdoe", owner=user1)
        team2 = Team.objects.create(name="teamdoe2", owner=user1)
        Team.objects.create(name="teamdoe3", owner=user1)
        Team.objects.create(name="teamdoe4", owner=user1)

        team1.members.add(user2)
        team2.members.add(user2, user3)

        # Create a public dataformat
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1,
            name='public_for_all',
            )
        assert dataformat, errors
        dataformat.share()

        self.setup_algorithms(declaration, code)


class AlgorithmsAPIBase(AlgorithmsTestCase):

    DECLARATION = """{
  "language": "python",
  "splittable": false,
  "groups": [
    {
      "name": "channel1",
      "inputs": {
        "a": { "type": "johndoe/single_integer/1" },
        "b": { "type": "johndoe/single_integer/1" }
      },
      "outputs": {
        "sum": { "type": "johndoe/single_integer/1" }
      }
    }
  ],
  "parameters": {
  }
}"""

    JSON_DECLARATION = json.loads(DECLARATION)

    CODE = """class Algorithm:

    def process(self, inputs, outputs):
        return True

"""

    UPDATE = """{
  "language": "python",
  "splittable": false,
  "groups": [
    {
      "name": "channel1",
      "inputs": {
        "a": { "type": "johndoe/single_integer/1" },
        "b": { "type": "johndoe/single_integer/1" }
      },
      "outputs": {
        "sum": { "type": "johndoe/single_integer/1" }
      }
    }
  ],
  "parameters": {
  }
}"""


    def setUp(self):
        super(AlgorithmsAPIBase, self).setUp()

        # Users
        user1 = User.objects.create_user('jackdoe', 'jackdoe@test.org', '1234')
        user2 = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        user3 = User.objects.create_user('janedoe', 'janedoe@test.org', '1234')
        User.objects.create_user('jacquelinedoe', 'jacquelinedoe@test.org', '1234')

        team1 = Team.objects.create(name="teamdoe", owner=user1)
        team2 = Team.objects.create(name="teamdoe2", owner=user1)
        Team.objects.create(name="teamdoe3", owner=user1)
        Team.objects.create(name="teamdoe4", owner=user1)

        team1.members.add(user2)
        team2.members.add(user3)

        # Create a public dataformat
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user2,
            name='single_integer',
            )
        assert dataformat, errors
        dataformat.share()

        self.setup_algorithms(self.DECLARATION, AlgorithmsAPIBase.CODE)

        # Private fork
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user2,
            name='forked_algo',
            short_description='',
            declaration=AlgorithmsAPIBase.DECLARATION,
            code=AlgorithmsAPIBase.CODE,
            )
        assert algorithm, errors

        algorithm.previous_version = None
        algorithm.fork_of = Algorithm.objects.get(name='public_for_all')
        algorithm.save()
