#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


from ..models import Algorithm

from .core import AlgorithmsAccessibilityFunctionsBase
from ...common.testutils import tearDownModule

class NotSharedAlgorithm_CheckAccessibilityFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_author_has_access(self):
        algorithm = Algorithm.objects.get(name='personal')

        errors = algorithm.is_accessible(public=False, users=['jackdoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_other_user_has_no_access(self):
        algorithm = Algorithm.objects.get(name='personal')

        errors = algorithm.is_accessible(public=False, users=['johndoe'])
        self.assertEqual(len(errors), 1)

        errors = algorithm.is_accessible(public=True, users=['johndoe'])
        self.assertEqual(len(errors), 1)


class UsableByOneUserAlgorithm_CheckAccessibilityFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_author_has_access(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        errors = algorithm.is_accessible(public=False, users=['jackdoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_authorized_user_has_usable_access(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        errors = algorithm.is_accessible(public=False, users=['johndoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['johndoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/usable_by_one_user/1' isn't accessible to the following users: johndoe")


    def test_unauthorized_user_has_no_access(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        errors = algorithm.is_accessible(public=False, users=['janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/usable_by_one_user/1' isn't usable by the following users: janedoe")

        errors = algorithm.is_accessible(public=True, users=['janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/usable_by_one_user/1' isn't accessible to the following users: janedoe")


    def test_mixed_case(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        errors = algorithm.is_accessible(public=False, users=['johndoe', 'janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/usable_by_one_user/1' isn't usable by the following users: janedoe")

        errors = algorithm.is_accessible(public=True, users=['johndoe', 'janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/usable_by_one_user/1' isn't accessible to the following users: johndoe, janedoe")


class UsableByAllUsersAlgorithm_CheckAccessibilityFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_author_has_access(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        errors = algorithm.is_accessible(public=False, users=['jackdoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_other_user_has_usable_access(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        errors = algorithm.is_accessible(public=False, users=['johndoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['johndoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/usable_by_all/1' isn't accessible to the following users: johndoe")


class PublicForOneUserAlgorithm_CheckAccessibilityFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_author_has_access(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        errors = algorithm.is_accessible(public=False, users=['jackdoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_authorized_user_has_full_access(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        errors = algorithm.is_accessible(public=False, users=['johndoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['johndoe'])
        self.assertEqual(len(errors), 0)


    def test_unauthorized_user_has_no_access(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        errors = algorithm.is_accessible(public=False, users=['janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/public_for_one_user/1' isn't usable by the following users: janedoe")

        errors = algorithm.is_accessible(public=True, users=['janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/public_for_one_user/1' isn't accessible to the following users: janedoe")


    def test_mixed_case(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        errors = algorithm.is_accessible(public=False, users=['johndoe', 'janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/public_for_one_user/1' isn't usable by the following users: janedoe")

        errors = algorithm.is_accessible(public=True, users=['johndoe', 'janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The algorithm 'jackdoe/public_for_one_user/1' isn't accessible to the following users: janedoe")


class PublicForAllUsersAlgorithm_CheckAccessibilityFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_author_has_access(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        errors = algorithm.is_accessible(public=False, users=['jackdoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_other_user_has_full_access(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        errors = algorithm.is_accessible(public=False, users=['johndoe'])
        self.assertEqual(len(errors), 0)

        errors = algorithm.is_accessible(public=True, users=['johndoe'])
        self.assertEqual(len(errors), 0)


class NotSharedAlgorithm_ShareFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_all(self):
        algorithm = Algorithm.objects.get(name='personal')

        algorithm.share(public=False)

        reference = {
            'status': 'usable',
            'shared_with': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_one_user(self):
        algorithm = Algorithm.objects.get(name='personal')

        algorithm.share(public=False, users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': [],
            'usable_by': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_multiple_users(self):
        algorithm = Algorithm.objects.get(name='personal')

        algorithm.share(public=False, users=['johndoe', 'janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': [],
            'usable_by': ['johndoe', 'janedoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_all(self):
        algorithm = Algorithm.objects.get(name='personal')

        algorithm.share(public=True)

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_one_user(self):
        algorithm = Algorithm.objects.get(name='personal')

        algorithm.share(public=True, users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'usable_by': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_multiple_users(self):
        algorithm = Algorithm.objects.get(name='personal')

        algorithm.share(public=True, users=['johndoe', 'janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
            'usable_by': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


class UsableByOneUserAlgorithm_ShareFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_all(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=False)

        reference = {
            'status': 'usable',
            'shared_with': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_the_same_user(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=False, users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': [],
            'usable_by': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_another_user(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=False, users=['janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': [],
            'usable_by': ['johndoe', 'janedoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_multiple_users(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=False, users=['janedoe', 'jacquelinedoe'])

        reference = {
            'status': 'shared',
            'shared_with': [],
            'usable_by': ['johndoe', 'janedoe', 'jacquelinedoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_all(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=True)

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_the_same_user(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=True, users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'usable_by': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_another_user(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=True, users=['janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['janedoe'],
            'usable_by': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_multiple_users(self):
        algorithm = Algorithm.objects.get(name='usable_by_one_user')

        algorithm.share(public=True, users=['janedoe', 'jacquelinedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['janedoe', 'jacquelinedoe'],
            'usable_by': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


class UsableByAllAlgorithm_ShareFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_all(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        algorithm.share(public=False)

        reference = {
            'status': 'usable',
            'shared_with': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_one_user(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        algorithm.share(public=False, users=['johndoe'])

        reference = {
            'status': 'usable',
            'shared_with': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_multiple_users(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        algorithm.share(public=False, users=['johndoe', 'janedoe'])

        reference = {
            'status': 'usable',
            'shared_with': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_all(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        algorithm.share(public=True)

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_one_user(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        algorithm.share(public=True, users=['johndoe'])

        reference = {
            'status': 'usable',
            'shared_with': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_multiple_users(self):
        algorithm = Algorithm.objects.get(name='usable_by_all')

        algorithm.share(public=True, users=['johndoe', 'janedoe'])

        reference = {
            'status': 'usable',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


class PublicForOneUserAlgorithm_ShareFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_all(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=False)

        reference = {
            'status': 'usable',
            'shared_with': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_the_same_user(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=False, users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'usable_by': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_another_user(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=False, users=['janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'usable_by': ['janedoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_multiple_users(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=False, users=['janedoe', 'jacquelinedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'usable_by': ['janedoe', 'jacquelinedoe'],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_all(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=True)

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_the_same_user(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=True, users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'usable_by': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_another_user(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=True, users=['janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
            'usable_by': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_multiple_users(self):
        algorithm = Algorithm.objects.get(name='public_for_one_user')

        algorithm.share(public=True, users=['janedoe', 'jacquelinedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe', 'jacquelinedoe'],
            'usable_by': [],
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


class PublicForAllAlgorithm_ShareFunction(AlgorithmsAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_all(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        algorithm.share(public=False)

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_one_user(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        algorithm.share(public=False, users=['johndoe'])

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_usable_by_multiple_users(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        algorithm.share(public=False, users=['johndoe', 'janedoe'])

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_all(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        algorithm.share(public=True)

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_one_user(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        algorithm.share(public=True, users=['johndoe'])

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)


    def test_sharing_as_public_for_multiple_users(self):
        algorithm = Algorithm.objects.get(name='public_for_all')

        algorithm.share(public=True, users=['johndoe', 'janedoe'])

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(algorithm, reference)
