#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.http import Http404, HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.contrib.auth.models import User

from .models import Algorithm
from ..team.models import Team
from ..common.texts import Messages
from ..ui.templatetags.markup import restructuredtext

from beat.core import prototypes

import simplejson as json

#----------------------------------------------------------


@login_required
def create(request, name=None):
    """Creates a new algorithm or a new version of an existing algorithm

    The user must be authenticated before it can add a new algorithm
    """

    parameters = {'algorithm_author':  request.user.username,
                  'algorithm_name':    name,
                  'algorithm_version': 1,
                  'short_description': '',
                  'description':       '',
                  'messages':          Messages,
                  'edition':           False,
                  'plot_account':      settings.PLOT_ACCOUNT,
                 }

    # Retrieves the existing algorithm (if necessary)
    if name is not None:
        previous_versions = Algorithm.objects.filter(
                                  author=request.user,
                                  name__iexact=name,
                              ).order_by('-version')
        if len(previous_versions) == 0:
            raise Http404()

        previous_version = previous_versions[0]

        description = previous_version.description

        parameters['algorithm_version'] = previous_version.version + 1
        parameters['source_code']       = previous_version.source_code_file.read()
        parameters['declaration']       = previous_version.declaration_file.read().replace('\n', '')
        parameters['short_description'] = previous_version.short_description
        parameters['description']       = description.replace('\n', '\\n')
        parameters['html_description']  = restructuredtext(description).replace('\n', '')
    else:
        parameters['source_code'] = prototypes.binary_load('algorithm.py')  \
                                           .replace('\n\n        # TODO: Implement this algorithm\n\n',
                                                    '\n        # TODO: Implement this algorithm\n')
        declaration, errors = prototypes.load('algorithm')
        declaration['language'] = 'python'
        parameters['declaration'] = json.dumps(declaration)

    return render_to_response('algorithms/edition.html',
                              parameters,
                              context_instance=RequestContext(request))


#----------------------------------------------------------


@login_required
def fork(request, author, name, version):
    """Creates a new algorithm by forking an existing algorithm

    The user must be authenticated before it can fork an algorithm
    """

    # Retrieves the forked algorithm based on user accessible algorithms
    fork_of = get_object_or_404(Algorithm.objects.for_user(request.user, True),
                                author__username__iexact=author,
                                name__iexact=name,
                                version=int(version)
                               )

    # The only case a user can't fork an algorithm is if it's usable rather
    # than shared
    if fork_of.sharing == Algorithm.USABLE and fork_of.author != request.user:
      raise Http404()

    description = fork_of.description

    parameters = {'original_author':   author,
                  'algorithm_author':  request.user.username,
                  'algorithm_name':    name,
                  'algorithm_version': fork_of.version,
                  'source_code':       fork_of.source_code_file.read(),
                  'declaration':       fork_of.declaration_file.read().replace('\n', ''),
                  'short_description': fork_of.short_description,
                  'description':       description.replace('\n', '\\n'),
                  'html_description':  restructuredtext(description).replace('\n', ''),
                  'messages':          Messages,
                  'fork_of':           fork_of,
                  'edition':           False,
                  'plot_account':      settings.PLOT_ACCOUNT,
                 }

    return render_to_response('algorithms/edition.html',
                              parameters,
                              context_instance=RequestContext(request))


#----------------------------------------------------------


@login_required
def edit(request, author, name, version):
    """Allows the modification of the algorithm

    The user must be authenticated before it can edit an algorithm
    """

    if author != request.user.username:
        raise Http404()

    # Retrieves the algorithm
    algorithm = get_object_or_404(Algorithm,
                                  author__username__iexact=author,
                                  name__iexact=name,
                                  version=int(version)
                                 )

    if not algorithm.modifiable():
        return HttpResponseForbidden('Algorithm %s is not modifiable' % \
                algorithm.fullname())

    description = algorithm.description

    parameters = {'algorithm_author':  request.user.username,
                  'algorithm_name':    name,
                  'algorithm_version': algorithm.version,
                  'source_code':       algorithm.source_code_file.read(),
                  'declaration':       algorithm.declaration_file.read().replace('\n', ''),
                  'short_description': algorithm.short_description,
                  'description':       description.replace('\n', '\\n'),
                  'html_description':  restructuredtext(description).replace('\n', ''),
                  'messages':          Messages,
                  'edition':           True,
                  'plot_account':      settings.PLOT_ACCOUNT,
                 }

    return render_to_response('algorithms/edition.html',
                              parameters,
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def view(request, author, name, version=None):
    """Shows the algorithm. The Web API is used to retrieve the details about
    the algorithm and check the accessibility.
    """

    # Retrieves the algorithm
    if version:
        algorithm = get_object_or_404(
                Algorithm,
                author__username__iexact=author,
                name__iexact=name,
                version=int(version),
                )
    else:
        algorithm = Algorithm.objects.filter(author__username__iexact=author,
                name__iexact=name).order_by('-version')
        if not algorithm:
            raise Http404()
        else:
            algorithm = algorithm[0]

    (has_access, _, __) = algorithm.accessibility_for(request.user)

    if not has_access: raise Http404()

    owner = (request.user == algorithm.author)

    # Users the object can be shared with
    users = User.objects.exclude(username__in=settings.ACCOUNTS_TO_EXCLUDE_FROM_TEAMS).order_by('username')

    # Render the page
    return render_to_response('algorithms/view.html',
                              {
                                  'algorithm': algorithm,
                                  'owner': owner,
                                  'users': users,
                                  'teams': Team.objects.for_user(request.user, True)
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def diff(request, author1, name1, version1, author2, name2, version2):
    """Shows the difference between two algorithms. The Web API is used to
    retrieve the details about the algorithms and check the accessibility.
    """

    return render_to_response('algorithms/diff.html',
                              {'algorithm1_author':  author1,
                               'algorithm1_name':    name1,
                               'algorithm1_version': version1,
                               'algorithm2_author':  author2,
                               'algorithm2_name':    name2,
                               'algorithm2_version': version2,
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def ls(request, author_name):
    '''List all accessible algorithms to the request user'''

    if not author_name: return public_ls(request)

    # check that the user exists on the system
    author = get_object_or_404(User, username=author_name)

    # orders algorithms so that the latest information is displayed first
    objects = Algorithm.objects.from_author_and_public(request.user,
            author_name).order_by('-creation_date')
    objects = Algorithm.filter_latest_versions(objects)

    return render_to_response('algorithms/list.html',
            dict(
                objects=objects,
                author=author,
                owner=(request.user==author),
                ),
            context_instance=RequestContext(request),
            )


#----------------------------------------------------------


def public_ls(request):
    '''List all publicly accessible objects'''

    # orders so that objects with latest information are displayed first
    objects = Algorithm.objects.public().order_by('-creation_date')
    objects = Algorithm.filter_latest_versions(objects)

    return render_to_response('algorithms/list.html',
            dict(
                objects=objects,
                author=request.user, #anonymous
                owner=False,
                ),
            context_instance=RequestContext(request),
            )
