#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

'''Scheduling functions and utilities'''

import os
import sys
import time
import socket
import signal
import subprocess
import logging
logger = logging.getLogger(__name__)

import simplejson

from django.conf import settings
from django.db import transaction

from .models import Job, JobSplit, Queue, Worker, Result


def _select_splits_for_queue(queue):
    '''Returns a list of job splits that can run now, at a certain queue

    Here is the work done:

    1. Find the queue availability. This is a bit tricky as queues are only
       allowed to consume a limited (configurable) number of slots in each
       worker, per user

    2. Calculate runnable job splits

    3. TODO: Calculates the list of job splits that can potentially run now
       (for which there is space in the current queue being analyzed), taking
       into consideration the relative use for every user and giving more
       priority to user with less load

    4. Return such a list clipping it so that the number of job splits returned
       does not exceed the queue availability

    The work is done inside the "queue" domain: it does not take into
    consideration other job splits that may also have slots shared on one of
    the machines this queue also has slots on.

    '''

    splits = JobSplit.objects.filter(job__block__queue=queue,
        worker__isnull=True, job__split_errors=0,
        status=Job.QUEUED).order_by('job__runnable_date')

    # lists of all users currently running
    users = set(splits.values_list('job__block__experiment__author', flat=True))

    # number of splits (== slots) running on this queue for each user
    user_slots = [JobSplit.objects.filter(job__block__experiment__author=k,
      job__block__queue=queue, status=Job.PROCESSING).count() for k in users]

    allowance = [queue.max_slots_per_user - k for k in user_slots]
    allowance = dict(zip(users, allowance))

    # limit runnable splits so we reach a maximum of allowed user slots
    splits_to_consider = []
    for s in splits:
        author = s.job.block.experiment.author.id
        if allowance[author] > 0:
            splits_to_consider.append(s)
            allowance[author] -= 1

    # TODO: Sort splits taking into consideration current user load

    # now we have a list of job splits, within each users' allowance, we could
    # submit, clip it by using the total number of available slots on the queue
    # the list is ordered having oldest job splits first
    return splits_to_consider[:queue.availability()]


@transaction.atomic
def send_experiment_emails():
    '''Sends experiment e-mails for experiments that just finished'''

    from ..experiments.models import Experiment

    qs = Experiment.objects.select_for_update().filter(email_dispatch=True)
    for k in qs: k.email()
    qs.update(email_dispatch=False)


@transaction.atomic
def schedule():
    '''Schedules job splits that can run now, respecting user/queue usage

    The priorities are:

    1. job splits that require more cores to run (more resource intense)
    2. job splits that are older (runnable since a long time)

    Reasoning: Jobs which require more cores are prioritary over job splits
    that don't require that many resources. The reason for this is practical:
    nodes associated to queues with more cores maybe used by queues that
    requires less slots in case no job splits requiring more cores are present,
    which will happen most of the time if there are more slots available with
    less resources. This will not work if either priorities are not established
    for the use of each node or you only have 1 machine hosting all queues. In
    these cases, you may face a situation in which job splits requesting more
    cores block the execution of job splits requesting less cores because of
    the lack of farm avalability.

    For example, a good setup could be like this (note priorities in this case
    are not useful, so they are all set to 0):

    Queue A (1 core/4Gb RAM/60 minutes/1 job per user):
      - Host 1 (1 core/4Gb): 1 slot (priority 0)
      - Host 2 (2 cores/8Gb): 2 slots (priority 0)

    Queue B (2 cores/8Gb RAM/120 minutes/1 job per user):
      - Host 2 (2 cores/8Gb): 1 slot (priority 0)

    In this way: job splits for Queue B have priority of job splits for Queue
    A.

    If no job splits are scheduled requiring 2 cores, then we have 3 slots
    available to run simpler job splits. In case a job with 2 cores pops-up,
    then we must make sure 1-core job splits running at ``Host 2`` are freed so
    to run that job. To do that, we must first check on more resource intensive
    job splits and then order those by age, processing one after the other.

    You may further setup a queue for "special" users that can run more job
    splits in parallel, like this:

    Queue A (1 core/4Gb RAM/60 minutes/1 job per user):
      - Host 1 (1 core/4Gb): 1 slot (priority 0)
      - Host 2 (2 cores/8Gb): 2 slots (priority 0)

    Queue A* (1 core/4Gb RAM/60 minutes/1 job per user):
      - Host 1 (1 core/4Gb): 1 slot (priority 0)
      - Host 2 (2 cores/8Gb): 2 slots (priority 1)

    Queue B (2 cores/8Gb RAM/120 minutes/1 job per user):
      - Host 2 (2 cores/8Gb): 1 slot (priority 2)

    Queue A* may be setup so that some "special" users can run more job splits
    than others, having the same configuration as Queue A, except for the
    priority on Host 2, which is higher. Users that have access to Queue A or
    Queue A* can use Host 1 on a "first come first served" basis (as they have
    the same priority). Host 2 is primarily used for 2-core job splits. If job
    splits requiring a single core appears to run on Host 2 (from either Queue
    A or A*), then job splits from Queue A* have priority.

    Here is a summary of the work done:

    1. Use :py:func:`_select_splits_for_queue` to figure out which splits can
       actually run on a queue basis, respecting user limits

    2. For those splits, prioritize execution by: number of cores required per
       job split and then, by age.

    3. Remove splits, based on the order, which have the same output

    4. Attribute more resource intensive job splits to the slots that are free.
       If no slots are free, then virtually block free slots on hosts that can
       process job splits requiring more resources. This will guarantee the
       remaining cores will be freed in the future and this resource-intensive
       job splits will be able to run at a certain point.

    5. For each job split on the list, check the queue virtual availability
       (taking into consideration blocking as executed on step 3) and leave the
       job split on the list if there is a free slot, otherwise remove it.


    Returns:

      list: The list of splits assigned at this scheduling iteration

    '''

    # updates jobs with split errors, cancel experiments if problems occur
    for j in Job.objects.select_for_update().filter(split_errors__gt=0):
        j._split_indices()

    # get queues in a good order
    sorted_queues = Queue.objects.order_by('-cores_per_slot',
        'max_slots_per_user')

    logger.debug('Ordered queues: %s',
        ','.join([k.name for k in sorted_queues]))

    splits_to_consider = \
        [_select_splits_for_queue(q) for q in sorted_queues]

    # decides which split to run considering the 'on-the-fly' availability

    # workers that can run job splits
    whitelist = {}
    for worker in Worker.objects.select_for_update().filter(active=True):
        availability = worker.available_cores()
        if availability <= 0: continue
        whitelist[worker] = availability

    assigned_splits = []

    for batch in splits_to_consider:

        if not batch: continue #empty list
        candidates = batch[0].job.block.queue.worker_availability()

        for split in batch:

            assigned = False
            required_cores = split.job.block.queue.cores_per_slot

            for c in candidates:
                avail = whitelist.get(c, 0)
                if not avail: continue #should not use this worker
                if avail >= required_cores:
                    logger.info("Assigning `%s' to worker `%s'", split, c)
                    split.schedule(c) #assign job split to worker
                    assigned_splits.append(split)
                    whitelist[c] -= required_cores
                    logger.debug("`%s' cores available: %d", c, whitelist[c])
                    assigned = True
                    break

            if not assigned and required_cores > 1:
                #blacklist the most promising worker, otherwise will never be
                #picked
                c0 = candidates[0]
                if c0 in whitelist:
                    logger.info("Could not assign `%s' so blacklisting `%s'",
                        split, c0)
                    del whitelist[c0]

    return assigned_splits


@transaction.atomic
def _atomic_split_end(split_pk, result):
    '''Tries to, atomically end the split by calling split.end()'''

    try:
        split = JobSplit.objects.select_for_update().get(pk=split_pk)
        split.end(result)
    except JobSplit.DoesNotExist:
        logger.warn("Job split(pk=%d) does not exist. Likely cancelled, " \
            "so ignoring result `%s'", split_pk, result)


def process(split_pk, execute, cpulimit=None, cache=settings.CACHE_ROOT):
    '''Process assigned job splits using beat.core

    This task executes the user algorithm on a subprocess. It also serves the
    data to the user process so it works like an I/O daemon.

    If ``required_slots == 1``, then this job takes care of the whole data set.
    Otherwise, it takes care of a subset of the input data that is synchronised
    with this block, determined by ``split_index``.

    Two processes are spawned from the current work process:

    * The process for executing the user code
    * A process to limit the CPU usage (with ``cpulimit``), if these conditions
      are respected:

      1. The program ``cpulimit`` is available on the current machine
      2. The configuration requests a CPU usage greater than 0 (``nb_cores >
         0``). (N.B.: a value of zero means not to limit on CPU).


    Parameters:

      split_pk (int): The primary-key of the JobSplit to process

      execute (str): The path to the ``execute`` program to use for running the
        user code associated with this job split.

      cpulimit (str, Optional): The path to the ``cpulimit`` program to use for
        limiting the user code in CPU usage. If not set, then don't use it,
        even if the select user queue has limits.

      cache (str, Optional): The path leading to the root of the cache to use
        for this run. If not set, use the global default at
        ``settings.CACHE_ROOT``.

    '''

    with transaction.atomic():
        # lock split
        try:
            split = JobSplit.objects.select_for_update().get(pk=split_pk)
            split.start()
        except JobSplit.DoesNotExist:
            logger.info("Job split(pk=%d) does not exist. Likely cancelled, " \
                "so, ignoring.", split_pk)
            return

    config = simplejson.loads(split.job.block.command)

    # setup range if necessary
    if split.job.block.required_slots > 1:

        if (split.start_index) is None or (split.end_index is None):
            message = "The split %d/%d (pid=%d) running on worker `%s' for " \
                "block `%s' of experiment `%s' could not " \
                "be completed: indexes are missing!" % \
                (split.split_index+1, split.job.block.required_slots,
                  split.process_id, split.worker, split.job.block.name,
                  split.job.block.experiment.fullname())
            logger.warn(message)
            _atomic_split_end(split_pk, Result(status=1, syserr=message,
                usrerr=settings.DEFAULT_USER_ERROR))

        config['range'] = [split.start_index, split.end_index]

    # For reference, this bit of code should match (or be very similar) to the
    # one at beat.cmdline.experiments:run_experiment()
    split = JobSplit.objects.get(pk=split_pk) #notice: no locking
    from beat.core.execution import Executor

    try:

        executor = Executor(settings.PREFIX, config, cache)

        if not executor.valid:
            err = ''
            for e in executor.errors: err += '  * %s\n' % e
            message = "Failed to load execution information for split %d/%d " \
                    "running at worker `%s', for block `%s' of experiment " \
                    "`%s': %s" % (split.split_index+1,
                        split.job.block.required_slots,
                        split.worker,
                        split.job.block.name,
                        split.job.block.experiment.fullname(),
                        err,
                        )
            raise RuntimeError(message)

        queue = split.job.block.queue
        nb_cores = queue.cores_per_slot
        if (nb_cores > 0) and (cpulimit is None):
            logger.warn("Job requires limiting CPU usage to %g (cores), but " \
                "you have not set the path to the program `cpulimit'. " \
                "Continuing without CPU limiting...", nb_cores)
            nb_cores = 0

        logger.info("Running `%s' on worker request", executor.algorithm.name)

        # n.b.: with executor may crash on the database view setup
        with executor:
            result = executor.process(
                    execute_path=execute,
                    virtual_memory_in_megabytes=queue.memory_limit,
                    max_cpu_percent=int(100*float(nb_cores)), #allows for 150%
                    cpulimit_path=cpulimit,
                    timeout_in_minutes=queue.time_limit,
                    daemon=0,
                    )

        _atomic_split_end(split_pk, Result(
            status=result['status'],
            stdout=result['stdout'],
            stderr=result['stderr'],
            usrerr=result['user_error'],
            syserr=result['system_error'],
            _stats=simplejson.dumps(result['statistics'], indent=2),
            ))
        logger.info("Process `%s' for split `%s' ended gracefully",
            split.process_id, split)

    except:
        from traceback import format_exc
        logger.warn("Process `%s' for split `%s' ended with an error: %s",
            split.process_id, split, format_exc())
        _atomic_split_end(split_pk, Result(status=1,
           usrerr=settings.DEFAULT_USER_ERROR, syserr=format_exc(),))


@transaction.atomic
def worker_update():
    '''Updates the worker state in the database from local readings'''

    # myself, raises if I cannot find me
    worker = Worker.objects.select_for_update().get(name=socket.gethostname())

    worker.update_state()


@transaction.atomic
def work(environments, cpulimit, process, django_settings):
    '''Launches user code on isolated processes

    This function is supposed to be called asynchronously, by a scheduling
    agent, every few seconds. It examines job splits assigned to the current
    host and launches an individual process to handle these splits. The process
    is started locally and the process ID stored with the split.

    Job split cancelling is executed by setting the split state as
    ``KILL`` and waiting for this function to handle it.


    Parameters:

      environments (dict): A dictionary containing installed environments,
        their description and execute-file paths.

      cpulimit (str): The path to the ``cpulimit`` program to use for
        limiting the user code in CPU usage. If set to ``None``, then don't use
        it, even if the select user queue has limits.

      process (str): The path to the ``process.py`` program to use for running
        the user code on isolated processes.

      django_settings (str): The name of the module containing the Django
        settings for use with the process program

    '''

    # myself, raises if I cannot find me
    worker = Worker.objects.get(name=socket.gethostname())

    # cancel job splits by killing associated processes
    for j in JobSplit.objects.select_for_update().filter(worker=worker,
        status=Job.KILL, end_date__isnull=True, process_id__isnull=False):
        os.kill(j.process_id, signal.SIGKILL)
        j.end(None, Job.CANCELLED)

    # cmdline base argument
    cmdline = [process, '--settings=%s' % django_settings]
    if cpulimit is not None: cmdline += ['--cpulimit=%s' % cpulimit]
    if settings.DEBUG: cmdline += ['-vv']

    # start newly assigned job splits
    for j in JobSplit.objects.select_for_update().filter(worker=worker,
        status=Job.QUEUED, start_date__isnull=True, process_id__isnull=True):
        execute = pick_execute(j, environments)
        if execute is None:
            message = "Environment `%s' is not available for split %d/%d " \
                    "running at worker `%s', for block `%s' of experiment " \
                    "`%s': %s" % (split.job.block.environment,
                        split.split_index+1,
                        split.job.block.required_slots,
                        split.worker,
                        split.job.block.name,
                        split.job.block.experiment.fullname(),
                        "Available environments are `%s'" % \
                            '|'.join(environments.keys()),
                        )
            j.end(Result(status=1, usrerr=settings.DEFAULT_USER_ERROR,
              syserr=message))

        # if we get to this point, then we launch the user process
        subprocess.Popen(cmdline + [execute, str(j.pk)])


def resolve_process_path():
  '''Returns the path to cpulimit'''

  basedir = os.path.dirname(os.path.realpath(sys.argv[0]))
  r = os.path.join(basedir, 'process.py')

  if not os.path.exists(r):
      raise RuntimeError("Cannot find `process.py' at `%s' - please check " \
          "your installation" % basedir)

  return r


def find_environments(paths=None):
  '''Finds list of known environments

  Parameters:

    paths (list, Optional): A list of paths where to search for environments.
      If not set, then load default environments if possible.


  Returns:

    dict: A dictionary containing each environment available using as key the
      natural key for environments (i.e., ``name (version)``) and as values
      another dictionary with these keys:

        * name: The environment name (str)
        * version: The environment version (str)
        * os: The output of ``uname -a`` (list):
          1. Operating system (str)
          2. Hostname (str)
          3. Kernel version (str)
          4. Kernel compilation details (str)
          5. Platform (``x86_64`` for 64-bits or ``i386`` for 32-bits) (str)
        * execute: The path to the ``execute`` script to be used for running
          user jobs (str)
        * directory: The path leading to the root of this environment (str)

  '''

  from beat.core.execution import discover_environments

  if paths is not None:
    logger.debug("Search for environments at `%s'", os.pathsep.join(paths))
    return discover_environments(paths)

  else:
    import pkg_resources
    path = pkg_resources.resource_filename(__name__, 'environments')
    logger.debug("Search for environments at `%s'", path)
    return discover_environments([path])


@transaction.atomic
def worker_shutdown():
    """Standard worker shutdown procedure

    Stop all running/assigned splits and then mark the worker as inactive.
    """

    # myself, raises if I cannot find me
    worker = Worker.objects.select_for_update().get(name=socket.gethostname())

    message = 'Cancelled on forced worker shutdown (maintenance)' \
        ' - you may retry submitting your experiment shortly'

    # cancel job splits which are running
    for j in JobSplit.objects.select_for_update().filter(worker=worker,
        status=(Job.KILL, Job.PROCESSING), end_date__isnull=True,
        process_id__isnull=False):
        signal.signal(signal.SIGKILL, j.process_id)
        j.end(Result(status=1, usrerr=message))

    # cancel job splits which were not yet started
    for j in JobSplit.objects.select_for_update().filter(worker=worker,
        status=Job.QUEUED, start_date__isnull=True, process_id__isnull=True):
        j.end(Result(status=1, usrerr=message))

    # lock worker and modify it
    worker.active = False
    worker.used_cores = 0
    worker.used_memory = 0
    worker.info = 'Worker shutdown by system administrator'
    worker.save()



def pick_execute(split, environments):
    """Resolves the path to the ``execute`` program to use for the split"""

    # Check we have a compatible environment to execute the user algorithm
    envinfo = environments.get(split.job.block.environment.natural_key())
    return envinfo['execute'] if envinfo else None
