#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from rest_framework import serializers

from ..common.serializers import ContributionCreationSerializer
from ..common.serializers import SharingSerializer
from ..common.serializers import ContributionSerializer
from ..common.serializers import DiffSerializer

from .models import Code

import simplejson as json
import difflib


#----------------------------------------------------------


class CodeCreationSerializer(ContributionCreationSerializer):
    code = serializers.CharField(required=False, allow_blank=True, trim_whitespace=False)

    class Meta(ContributionCreationSerializer.Meta):
        fields = ContributionCreationSerializer.Meta.fields + ['code']


#----------------------------------------------------------


class CodeSharingSerializer(SharingSerializer):
    status = serializers.CharField()

    def validate_status(self, value):
        if value not in ['public', 'usable']:
            raise serializers.ValidationError('Invalid status value')
        return value


#----------------------------------------------------------


class CodeSerializer(ContributionSerializer):
    opensource = serializers.SerializerMethodField()
    modifiable = serializers.BooleanField()

    class Meta(ContributionSerializer.Meta):
        model = Code
        default_fields = ContributionSerializer.Meta.default_fields + ['opensource']
        extra_fields = ContributionSerializer.Meta.extra_fields + ['code']
        exclude = ContributionSerializer.Meta.exclude + ['source_code_file']

    def __init__(self, *args, **kwargs):
        # Don't pass the 'opensource' arg up to the superclass
        self.opensource = kwargs.pop('opensource', False)

        # Instantiate the superclass normally
        super(ContributionSerializer, self).__init__(*args, **kwargs)

    def get_opensource(self, obj):
        user = self.context.get('user')
        (has_access, open_source, accessibility) = obj.accessibility_for(user)
        return open_source

    def get_accessibility(self, obj):
        if obj.sharing == Code.PUBLIC:
            return 'public'
        elif obj.sharing == Code.SHARED or obj.sharing == Code.USABLE:
            return 'confidential'
        else:
            return 'private'

    def get_sharing(self, obj):
        user = self.context.get('user')

        sharing = super(CodeSerializer, self).get_sharing(obj)
        if user == obj.author:
            if obj.usable_by.count() > 0:
                sharing['usable_by'] = map(lambda x: x.username,
                                                     obj.usable_by.iterator())

            if obj.usable_by_team.count() > 0:
                sharing['usable_by_team'] = map(lambda x: x.name,
                                                     obj.usable_by_team.iterator())
        return sharing

    def get_code(self, obj):
        user = self.context.get('user')

        if obj.author != user:
            (has_access, open_source, accessibility) = obj.accessibility_for(user)
            if not(has_access) or not(open_source):
                return None

        return obj.source_code


#----------------------------------------------------------


class CodeDiffSerializer(DiffSerializer):
    source_code_diff = serializers.SerializerMethodField()

    def get_source_code_diff(self, obj):
        source1 = obj['object1'].source_code
        source2 = obj['object2'].source_code
        diff = difflib.ndiff(source1.splitlines(), source2.splitlines())

        return '\n'.join(filter(lambda x: x[0] != '?', list(diff)))
