#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


from django.db import models
from django.dispatch import receiver

from ..dataformats.models import DataFormat

from .models import Database, DatabaseProtocol, DatabaseSet
from .models import DatabaseSetTemplate, DatabaseSetTemplateOutput
from .models import DatabaseSetOutput

from .models import validate_database


@receiver(models.signals.post_delete, sender=Database)
def auto_delete_file_on_delete(sender, instance, **kwargs):
    """Deletes file from filesystem when ``Database`` object is deleted.
    """
    if instance.declaration_file:
        instance.declaration_file.delete(save=False)

    if instance.source_code_file:
        instance.source_code_file.delete(save=False)

    if instance.description_file:
        instance.description_file.delete(save=False)


@receiver(models.signals.pre_save, sender=Database)
def auto_delete_file_on_change(sender, instance, **kwargs):
    """Deletes file from filesystem when ``Database`` object is changed."""

    if not instance.pk:
        return False

    try:
        old_file = Database.objects.get(pk=instance.pk).declaration_file
    except Database.DoesNotExist:
        return False

    if old_file != instance.declaration_file:
        old_file.delete(save=False)

    try:
        old_code = Database.objects.get(pk=instance.pk).source_code_file
    except Database.DoesNotExist:
        return False

    if old_code != instance.source_code_file:
        old_code.delete(save=False)

    try:
        old_descr = Database.objects.get(pk=instance.pk).description_file
    except Database.DoesNotExist:
        return False

    if old_descr != instance.description_file:
        old_descr.delete(save=False)


@receiver(models.signals.post_save, sender=Database)
def refresh_protocols(sender, instance, **kwargs):
    """Refreshes changed protocols"""

    try:

        core = validate_database(instance.declaration)
        core.name = instance.fullname()

        protocols = DatabaseProtocol.objects.filter(
                database__name=instance.name,
                database__version=instance.version,
            )

        existing = set((k.name, k.set_template_basename()) for k in protocols)
        new_objects = set([(k,v['template']) for k,v in core.protocols.items()])

        for protocol_name, template in existing - new_objects:
            # notice: no need to worry, this will clean-up all the rest
            protocols.get(name__iexact=protocol_name).delete()

        json_protocols = dict([(k,v) for k,v in core.protocols.items()])

        for protocol_name, template in new_objects - existing:
            protocol = DatabaseProtocol(name=protocol_name, database=instance)
            protocol.save()

            json_protocol = json_protocols[protocol_name]

            # creates all the template sets, outputs, etc for the first time
            for set_attr in json_protocol['sets']:

                tset_name = json_protocol['template'] + '__' + set_attr['template']

                dataset_template = DatabaseSetTemplate.objects.filter(name=tset_name)
                if not dataset_template: #create
                    dataset_template = DatabaseSetTemplate(name=tset_name)
                    dataset_template.save()
                else:
                    dataset_template = dataset_template[0]

                # Create the database set
                dataset = DatabaseSet.objects.filter(
                    name = set_attr['name'],
                    template = dataset_template,
                    protocol = protocol,
                    )

                if not dataset: #create
                    dataset = DatabaseSet(
                        name = set_attr['name'],
                        template = dataset_template,
                        protocol = protocol,
                        )
                    dataset.save()

                # Create the database set template output
                for output_name, format_name in set_attr['outputs'].items():
                    if len(format_name.split('/')) != 3:
                        raise SyntaxError(
                            "Dataformat should be named following the " \
                            "style `<username>/<format>/<version>', the " \
                            "value `%s' is not valid" % (format_name,)
                            )
                    (author, name, version) = format_name.split('/')
                    dataformats = DataFormat.objects.filter(
                        author__username=author,
                        name=name,
                        version=version,
                        )

                    # TODO: Remove this when validation works (see comments)
                    if len(dataformats) != 1:
                        raise SyntaxError(
                            "Could not find dataformat named `%s' to set" \
                            "output `%s' of template `%s' for protocol" \
                            "`%s' of database `%s'", (
                              format_name,
                              output_name,
                              dataset_template.name,
                              protocol_name,
                              instance.name,
                              )
                            )
                        return

                    database_template_output = \
                        DatabaseSetTemplateOutput.objects.filter(
                            name=output_name,
                            template=dataset_template,
                            dataformat=dataformats[0],
                            )

                    if not database_template_output: # create
                        database_template_output = \
                            DatabaseSetTemplateOutput(
                                name=output_name,
                                template=dataset_template,
                                dataformat=dataformats[0],
                                )
                        database_template_output.save()

                    else:
                        database_template_output = \
                            database_template_output[0]

                    # Create the database set output
                    hash = core.hash_output(protocol.name,
                        dataset.name, output_name)
                    dataset_output = \
                        DatabaseSetOutput.objects.filter(hash=hash)

                    if not dataset_output: # create
                        dataset_output = DatabaseSetOutput(
                            template=database_template_output,
                            set=dataset,
                            hash=hash,
                            )
                        dataset_output.save()

    except Exception:
        instance.delete()
        raise


@receiver(models.signals.post_delete, sender=DatabaseSet)
def delete_empty_template_sets(sender, **kwargs):

    instance = kwargs['instance']
    try:
        if not instance.template.sets.all(): instance.template.delete()
    except:
        pass
