#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################



from django import template
from django.conf import settings
from django.db.models import Count
from django.db.models import Q

import simplejson as json

from ..models import Result
from ...plotters.models import Plotter


register = template.Library()


#--------------------------------------------------


@register.inclusion_tag('experiments/panels/table.html', takes_context=True)
def experiment_table(context, objects, owner, id):
    return dict(
            request=context['request'],
            objects=objects,
            owner=owner,
            panel_id=id,
            )


#--------------------------------------------------


@register.inclusion_tag('experiments/panels/breadcrumb.html', takes_context=True)
def experiment_breadcrumb(context, obj):
    return dict(
            request=context['request'],
            object=obj,
            )


#--------------------------------------------------


@register.inclusion_tag('experiments/panels/filter_script.html')
def filter_script(panel_id, text_filter_id, attestation_filter_id, privacy_filter_id, status_filter_id):
    return dict(
            panel_id=panel_id,
            text_filter_id=text_filter_id,
            attestation_filter_id=attestation_filter_id,
            privacy_filter_id=privacy_filter_id,
            status_filter_id=status_filter_id,
            )


#--------------------------------------------------


@register.inclusion_tag('experiments/panels/report_script.html',
        takes_context=True)
def report_script(context, panel_id, master_checkbox_id, checkbox_class,
        form_id):
    return dict(
            request=context['request'],
            panel_id=panel_id,
            master_checkbox_id=master_checkbox_id,
            checkbox_class=checkbox_class,
            form_id=form_id,
            )


#--------------------------------------------------


@register.inclusion_tag('experiments/panels/execution_infos.html')
def experiment_execution_infos(id):
    return dict(
            panel_id=id,
            URL_PREFIX=settings.URL_PREFIX,
            )


#--------------------------------------------------


@register.inclusion_tag('experiments/dialogs/algorithm_mapping.html')
def algorithm_mapping(id):
    return dict(
            dialog_id=id,
            URL_PREFIX=settings.URL_PREFIX,
            )


#--------------------------------------------------


@register.inclusion_tag('experiments/panels/actions.html', takes_context=True)
def experiment_actions(context, object, display_count):
    '''Composes the action buttons for a particular experiment

    This panel primarily exists for showing action buttons for a given
    experiment taking into consideration it is being displayed for a given
    user.

    Parameters:

        object (Experiment): The experiment object concerned for which the
          buttons will be drawn.
        display_count (bool): If the set of buttons should include one with the
          number of forkes based on this experiment.

    '''
    return dict(
            request=context['request'],
            object=object,
            display_count=display_count,
            )


@register.inclusion_tag('experiments/panels/sharing.html', takes_context=True)
def experiment_sharing(context, obj):
    '''Composes the current sharing properties and a form to change them

    Parameters:

        object (Toolchain): The experiment concerned for which the buttons will
          be drawn.

    '''
    return {
            'request': context['request'],
            'object': obj,
            'users': context['users'],
            'teams': context['teams'],
            }


#--------------------------------------------------


@register.simple_tag
def experiment_list_of_referenced_formats(experiment):
    dataformats = filter(lambda x: x.author == experiment.author, experiment.all_needed_dataformats())

    return '[%s]' % ','.join(map(lambda x: '{ "name": "%s", "accessibility": "%s" }' % (x.fullname(), x.get_sharing_display().lower()),
                                 dataformats))


#--------------------------------------------------

@register.simple_tag
def experiment_list_of_referenced_algorithms(experiment, user):
    algorithms = filter(lambda x: x.author == experiment.author, experiment.referenced_algorithms.iterator())

    result = []
    for algorithm in algorithms:
        (has_access, open_source, accessibility) = algorithm.accessibility_for(user)
        if not(has_access):
            continue

        entry = {
            'name':          algorithm.fullname(),
            'accessibility': accessibility,
            'opensource':    open_source,
        }

        result.append(entry)

    return json.dumps(result)


#--------------------------------------------------


@register.inclusion_tag('experiments/panels/results.html')
def experiment_results(panel_id, experiment):
    """Called to display the experiment result panel.

    You must use this tag **with care**, as it will show experimental results
    even if the current user has no access to the experiment - this behaviour
    is useful for attestations and reports.
    """

    # create doubles of analyzer -> result pairs
    analyzers = experiment.blocks.filter(analyzer=True)
    contents = [(k, k.results.order_by('type', 'name')) for k in analyzers]

    # calculates plotter/parameter associations for the required types
    formats = [k.type for a in analyzers for k in a.results.exclude(type__in=Result.SIMPLE_TYPE_NAMES)]

    # get capable plotters and parameters associated
    plotting = dict([(k, Plotter.objects.for_strformat(k)) for k in formats])

    json_plotting = dict()
    for fmt in plotting:
        plotter, param = plotting[fmt]['default']
        parameters = plotting[fmt]['options'][plotter]
        json_plotting[fmt] = dict(
                plotter=plotter.fullname(),
                parameter=param.fullname() if param else '',
                parameters=json.dumps([k.fullname() for k in parameters]),
                )

    return dict(
            analyzers=contents,
            plotting=json_plotting,
            panel_id=panel_id,
            URL_PREFIX=settings.URL_PREFIX,
            )


#--------------------------------------------------


@register.assignment_tag
def ordered_blocks(xp):
    """Calculates the order in which blocks must be displayed"""

    block_order = xp.toolchain.core().execution_order().keys()
    blocks = []
    analyzers = []
    for key in block_order:
      b = xp.blocks.get(name=key)
      if b.analyzer: analyzers.append(b)
      else: blocks.append(b)
    return blocks + analyzers


#--------------------------------------------------


@register.assignment_tag(takes_context=True)
def visible_reports(context, xp):
    """Calculates the visible reports to an experiment"""
    from ...reports.models import Report
    user = context['request'].user
    if user.is_anonymous():
        return xp.reports.filter(status=Report.PUBLISHED)
    else:
        return xp.reports.filter(Q(author=context['request'].user) |
            Q(status=Report.PUBLISHED))


#--------------------------------------------------


@register.assignment_tag
def owner_algorithms(obj):
    '''Calculates the user algorithms for a given experiment'''

    return obj.referenced_algorithms.filter(author=obj.author)


@register.assignment_tag
def owner_dataformats(obj):
    '''Calculates the user dataformats and algorithms for a given attestation'''

    return [k for k in obj.all_needed_dataformats() if k.author == obj.author]
