#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import itertools
import base64
import collections

import logging
logger = logging.getLogger(__name__)

import simplejson

from django.http import HttpResponse, HttpResponseForbidden, HttpResponseBadRequest
from django.conf import settings
from django.shortcuts import get_object_or_404

from ..experiments.models import Experiment, Block, Result
from ..dataformats.models import DataFormat
from ..reports.models import Report

from .models import Plotter, PlotterParameter, DefaultPlotter

import beat.core.experiment
import beat.core.toolchain


def plot(request):
    """Plots one or more experiment analyzer outputs following the user
    preference and the plotter


    This view receives a single data payload with a GET or POST request
    containing multiple keys for ``experiment``, ``analyzer`` and ``output``.
    We'll cycle through (with :py:func:`itertools.cycle`) them to recover a
    dictionary like this::

         {
           'user/user/toolchain/1/label': {
             "analyzer": "my_analysis",
             "output" : "my_plot"
           },
           'user/user/toolchain/1/another': {
             "analyzer": "my_analysis",
             "output" : "my_plot"
           },
           'user/user/toolchain/3/xxx': {
             "analyzer": "the_greatest_analysis",
             "output" : "scatter"
           }
         }

    A single ``plotter``, if passed, should reference the plotter to use. It
    may be omitted, in which case we use the default plotter for the first
    experiment in the list.


    Returns:

      A response with a single image inside, following the plotter
      specifications.

    """

    if request.GET: use = request.GET
    elif request.POST: use = request.POST
    else:
        return HttpResponseBadRequest('This view must be called with either a GET or POST request')

    report_experiments = None

    if use.has_key("report_number"):
        try:
            report = Report.objects.get(number = int(use["report_number"]))
        except:
            message = "Invalid report number"
            logger.warn(message)
            return HttpResponseBadRequest(message)

        alias_filter =  use.getlist('experiment') or use.getlist('experiment[]')
        report_experiments, report_legends = report._get_experiments_and_alias(alias_filter)

    experiments = collections.OrderedDict()
    for xpid, analyzer, output in itertools.izip(
            report_experiments or use.getlist('experiment') or use.getlist('experiment[]'),
            itertools.cycle(use.getlist('analyzer') or use.getlist('analyzer[]')),
            itertools.cycle(use.getlist('output') or use.getlist('output[]')),
            ):
        experiments[xpid] = {
                'analyzer': analyzer,
                'output': output,
                }

    keywords = [
            'experiment',
            'experiment[]',
            'analyzer',
            'analyzer[]',
            'output',
            'output[]',
            'plotter',
            'parameter',
            'report_number',
            ]

    extra_parameters = {}
    for key in [k for k in use.keys() if k not in keywords]:
        extra_parameters[key] = '&'.join(use.getlist(key))

    # If no experiment to display, returns a 400
    if not experiments:
        message = "Plot at least one experiment's result"
        logger.warn(message)
        return HttpResponseBadRequest(message)

    # Retrieve the experiment
    for k in experiments:
        es = beat.core.experiment.Storage(settings.PREFIX, k)
        ts = beat.core.toolchain.Storage(settings.PREFIX, es.toolchain)
        try:
            xp = Experiment.objects.get(
                author__username=es.username,
                toolchain__author__username=ts.username,
                toolchain__name=ts.name,
                toolchain__version=ts.version,
                name=es.name,
                )
        except Experiment.DoesNotExist as e:
            message = "Experiment `%s' is not accessible" \
                    % '/'.join((es.username, ts.username, ts.name, ts.version,
                        es.name))
            logger.warn(message)
            return HttpResponseBadRequest(message)

        # Check that the user has access to the experiment or that an
        # attestation has been emitted for it (in which case somebody with the
        # URL may access the results/plots).
        has_access, _ = xp.accessibility_for(request.user)
        if not (has_access or xp.has_attestation):
            message = 'User `%s` not allowed access to experiment `%s`' % (request.user.username, xp.fullname())
            logger.warn(message)
            return HttpResponseForbidden(message)

        experiments[k]['obj'] = xp #save


    # Get the plotter and parameter, resolve if empty
    final_plotter = None #use the default
    user_parameter = None #use the default
    corefmt = None

    if 'plotter' in use:
        username, name, version = use['plotter'].split('/')
        try:
            final_plotter = Plotter.objects.get(author__username=username,
                  name=name, version=version)
        except Plotter.DoesNotExist as e:
            message = "Plotter `%s' is not accessible" % \
                    '/'.join((username, name, version))
            logger.warn(message)
            return HttpResponseBadRequest(message)

    if 'parameter' in use and use['parameter'].strip():
        username, name, version = use['parameter'].split('/')
        try:
            user_parameter = PlotterParameter.objects.get(
                    author__username=username,
                    name=name,
                    version=version,
                    )
        except PlotterParameter.DoesNotExist as e:
            message = "Parameter `%s' is not accessible" % \
                    '/'.join((username, name, version))
            logger.warn(message)
            return HttpResponseBadRequest(message)


    # Collect the data for the plot, check compatibility
    default = None
    for k, v in experiments.items():
        if v['analyzer'].find('/') >= 0:
            (analyzer_author, analyzer_name, analyzer_version) = v['analyzer'].split('/')
            xp = v['obj']
            block = get_object_or_404(Block,
                experiment=v['obj'],
                algorithm__author__username=analyzer_author,
                algorithm__name=analyzer_name,
                algorithm__version=analyzer_version,
                )
        else:
            block = get_object_or_404(Block,
                experiment=v['obj'],
                name=v['analyzer'],
                )

        result = get_object_or_404(Result,
            cache=block.outputs.get(),
            name=v['output'],
            )

        # now decide if the chart is compatible
        username, name, version = result.type.split('/')
        try:
            dfobj = DataFormat.objects.get(author__username=username,
                    name=name, version=int(version))
        except DataFormat.DoesNotExist as e:
            message = "Data format `%s' is not accessible" % \
                    '/'.join((username, name, version))
            logger.warn(message)
            return HttpResponseBadRequest(message)

        if final_plotter and final_plotter.dataformat != dfobj:
            message = 'Result with id=%d and type=%s is incompatible with plotter %s' % (result.id, result.type, final_plotter.fullname())
            logger.warn(message)
            return HttpResponseBadRequest(message)

        # check compatibility, fill up defaults
        if default is None:
            default = DefaultPlotter.objects.filter(dataformat=dfobj)

            if not default and not final_plotter:
                message = 'No plotter specified and no default for plot format %s' % dfobj.fullname()
                return HttpResponseBadRequest(message)

            default = default[0] #get the first and only

            # set defaults, if specific values have not already been set
            if not final_plotter: final_plotter = default.plotter

            if corefmt is None: #loads it once
                corefmt = final_plotter.core_format()

        # loads the data, for that particular result
        parsed = simplejson.loads(result.data_value)
        v['data'] = corefmt.type().from_dict(parsed, casting='unsafe')


    # checks the plotter is valid
    core_plotter = final_plotter.core()
    if not core_plotter.valid:
        message = 'Plotter %s is invalid' % final_plotter.fullname()
        return HttpResponseBadRequest(message)


    # resolves parameters, in order of priority
    final_parameters = {}
    if default.parameter: #fills-up defaults for the type
        final_parameters.update(simplejson.loads(default.parameter.data))
    if user_parameter: #fills-up defaults from the user set
        final_parameters.update(simplejson.loads(user_parameter.data))
    if extra_parameters: #fills-up defaults specified on the query
        final_parameters.update(extra_parameters)

    extra_parameters.setdefault('content_type', 'image/png') #in case not set

    # for HTML display, you need to set this correctly
    do_b64_encode = bool(extra_parameters.get('base64', False))

    # we filter out parameters we can't handle
    for key in final_parameters.keys():
        if key not in core_plotter.parameters: del final_parameters[key]


    runnable = core_plotter.runner()
    runnable.setup(final_parameters)
    assert runnable.ready
    data_to_plot = [(v['obj'].name,v['data']) for v in experiments.values()]
    fig = runnable.process(data_to_plot)
    if do_b64_encode: fig = base64.b64encode(fig)
    return HttpResponse(fig, content_type=final_parameters['content_type'])


#----------------------------------------------------------


def list_plotters(request, author_name):
    '''List all accessible plotters to the request user'''

    return render_to_response('plotters/list.html',
            dict(objects=Plotter.objects.from_author_and_public(request.user,
                author_name)),
            context_instance=RequestContext(request),
            )
