#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################



from django import template
from django.conf import settings
from ..models import Report


register = template.Library()


#--------------------------------------------------


@register.inclusion_tag('reports/panels/table.html', takes_context=True)
def report_table(context, objects, owner, id):
    '''Composes a report list table

    This panel primarily exists for user's report list page.

    Parameters:

        objects (iterable): An iterable containing Algorithm objects
        owner (bool): A flag indicating if the list is being created for the
          owner of his personal list page on the user-microsite or not.
        id: The HTML id to set on the generated table. This is handy for the
          filter functionality normally available on list pages.

    '''

    return dict(
            request=context['request'],
            objects=objects,
            owner=owner,
            panel_id=id,
            )


#--------------------------------------------------


@register.inclusion_tag('reports/panels/breadcrumb.html', takes_context=True)
def report_breadcrumb(context, obj):
    return {
            'request': context['request'],
            'owner': context['owner'],
            'object': obj,
           }


#--------------------------------------------------


@register.inclusion_tag('reports/panels/actions.html', takes_context=True)
def report_actions(context, object, display_count):
    '''Composes the action buttons for a particular report

    This panel primarily exists for showing action buttons for a given
    report taking into consideration it is being displayed for a given user.

    Parameters:

        object (Algorithm): The report object concerned for which the
          buttons will be drawn.
        display_count (bool): If the set of buttons should include one with the
          number of experiments using this report.

    '''
    return dict(
            request=context['request'],
            object=object,
            display_count=display_count,
            )


#--------------------------------------------------


@register.inclusion_tag('reports/panels/viewer.html', takes_context=True)
def report_viewer(context, object, owner):
    '''Composes the search results for visualization

    Parameters:


        object (Report): The search object concerned for which the
          buttons will be drawn.

        owner (bool): A flag that will indicate if it is the owner that is
          visualzing the search or not. This triggers a change in which panels
          are opened on the result page. Owner visualizations collapse filters
          and documentation, focusing on the results. Non-owner visualizations
          collapse only the filter panel.

    '''
    return dict(
            request=context['request'],
            object=object,
            owner=owner,
            URL_PREFIX=context['URL_PREFIX'],
            )


#--------------------------------------------------


@register.assignment_tag(takes_context=True)
def visible_experiments(context, object):
    '''Calculates the visible experiments for a given report and requestor'''

    return object.experiments.for_user(context['request'].user, True)


#--------------------------------------------------


def report_unlocking(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/unlocking.html')(report_unlocking)


#--------------------------------------------------


def report_add_experiments(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/report_add_experiments.html')(report_add_experiments)


#--------------------------------------------------


def report_creation(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/new_report_dialog.html')(report_creation)


#--------------------------------------------------


def report_deletion(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX,
             'contribution_label': 'report',
           }


register.inclusion_tag('ui/contribution_deletion.html')(report_deletion)


#--------------------------------------------------


def report_lock(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/report_lock.html')(report_lock)


#--------------------------------------------------


def report_publish(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/report_publish.html')(report_publish)


#--------------------------------------------------


def report_remove_experiment(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/report_remove_experiment.html')(report_remove_experiment)


#--------------------------------------------------


def report_remove_item(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/report_remove_item.html')(report_remove_item)


#--------------------------------------------------


def report_saved(id):
    return { 'dialog_id': id,
             'URL_PREFIX': settings.URL_PREFIX
           }


register.inclusion_tag('reports/dialogs/report_saved.html')(report_saved)
