#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.shortcuts import render_to_response
from django.shortcuts import get_object_or_404
from django.template import RequestContext, Context
from django.conf import settings
from django.views.generic import TemplateView
from django.contrib.auth.models import User
from django.db.models import Q
from django.db.models.functions import Coalesce

from ..ui.templatetags.markup import restructuredtext

from .models import Report

import simplejson as json


#------------------------------------------------


def by_number(request, number):

    # get the query from the DB
    obj = get_object_or_404(Report, number=int(number))

    return render_to_response('reports/report.html',
            {
                'report_number' : number,
                'owner': False,
                'report': obj,
                'USE_HTTPS_GRAVATAR': settings.USE_HTTPS_GRAVATAR,
                },
            context_instance=RequestContext(request))


#------------------------------------------------


def for_author(request, author_name, report_name):

    # get the query from the DB
    obj = get_object_or_404(Report,
            author__username = author_name,
            name = report_name)


    return render_to_response('reports/report.html',
            {
                'author'      : author_name,
                'report_name' : report_name,
                'owner'       : (request.user == obj.author),
                'report'      : obj,
                'USE_HTTPS_GRAVATAR': settings.USE_HTTPS_GRAVATAR,
                },
            context_instance=RequestContext(request))


#------------------------------------------------


def ls(request, author_name):
    '''List all accessible reports to the request user'''

    if not author_name: return public_ls(request)

    # check that the user exists on the system
    author = get_object_or_404(User, username=author_name)
    owner  = (request.user==author)

    if owner:
        # user wants to retrieve all available objects, including his
        objects = Report.objects.filter(Q(status=Report.PUBLISHED)|Q(author=author))
    else:
        # user wants to retrieve objects belonging to another user
        objects = Report.objects.filter(author=author, status=Report.PUBLISHED)

    objects = objects.annotate(updated=Coalesce('publication_date', 'creation_date',)).order_by('-updated')

    return render_to_response('reports/list.html',
            dict(
                objects=objects,
                author=author,
                owner=owner,
                ),
            context_instance=RequestContext(request),
            )


#------------------------------------------------


def public_ls(request):
    '''List all publicly accessible reports'''

    objects = Report.objects.filter(status=Report.PUBLISHED).order_by('publication_date')

    objects = objects.annotate(updated=Coalesce('publication_date', 'creation_date',)).order_by('-updated')

    return render_to_response('reports/list.html',
            dict(
                objects=objects,
                author=request.user,
                owner=False,
                ),
            context_instance=RequestContext(request),
            )


#------------------------------------------------


class PartialGroupView(TemplateView):
    def get_template_names(self):
        if 'template_name' in self.kwargs:
            self.template_name = 'reports/partials/' + self.kwargs.get('template_name')
        return super(PartialGroupView, self).get_template_names()

    def get_context_data(self, **kwargs):
        context = super(PartialGroupView, self).get_context_data(**kwargs)
        return context
