#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import re
import simplejson as json

from django import template
from django.conf import settings

from ...experiments.models import Result
from ...plotters.models import Plotter

register = template.Library()


@register.inclusion_tag('search/panels/table.html', takes_context=True)
def search_table(context, objects, owner, id):
    '''Composes a search list table

    This panel primarily exists for user's search list page.

    Parameters:

        objects (iterable): An iterable containing Search objects
        owner (bool): A flag indicating if the list is being created for the
          owner of his personal list page on the user-microsite or not.
        id: The HTML id to set on the generated table. This is handy for the
          filter functionality normally available on list pages.

    '''

    return dict(
            request=context['request'],
            objects=objects,
            owner=owner,
            panel_id=id,
            )


@register.inclusion_tag('search/panels/breadcrumb.html', takes_context=True)
def search_breadcrumb(context, obj):
    return {
            'request': context['request'],
            'object': obj,
           }


@register.inclusion_tag('search/panels/actions.html', takes_context=True)
def search_actions(context, object, display_count):
    '''Composes the action buttons for a particular search

    This panel primarily exists for showing action buttons for a given
    search taking into consideration it is being displayed for a given user.

    Parameters:

        object (Search): The search object concerned for which the
          buttons will be drawn.
        display_count (bool): If the set of buttons should include one with the
          number of experiments using this search.

    '''
    return dict(
            request=context['request'],
            object=object,
            display_count=display_count,
            )


@register.inclusion_tag('search/panels/sharing.html', takes_context=True)
def search_sharing(context, object):
    '''Composes the current sharing properties and a form to change them

    Parameters:

        object (Search): The search object concerned for which the
          buttons will be drawn.

    '''
    return dict(
            object=object,
            users=context['users'],
            teams=context['teams'],
            )


@register.inclusion_tag('search/panels/viewer.html', takes_context=True)
def search_viewer(context, object, results, filters, display_settings, owner):
    '''Composes the search results for visualization

    Parameters:


        object (Search): The search object concerned for which the
          buttons will be drawn.

        owner (bool): A flag that will indicate if it is the owner that is
          visualzing the search or not. This triggers a change in which panels
          are opened on the result page. Owner visualizations collapse filters
          and documentation, focusing on the results. Non-owner visualizations
          collapse only the filter panel.

    '''
    return dict(
            request=context['request'],
            filters=filters,
            settings=display_settings,
            object=object,
            results=results,
            owner=owner,
            URL_PREFIX=context['URL_PREFIX'],
            )


@register.inclusion_tag('search/panels/results.html', takes_context=True)
def results_table(context, object, results, id):
    '''Composes the search results for visualization

    Parameters:


        object (Search): The search object concerned for which the
          buttons will be drawn.

        results (dict): Containing the following fields:

          'experiments': [Experiment, ...]
          'common_toolchains': [Toolchain, ...],
          'common_protocols': [Protocol, ...],
          'common_analyzers': [(Algorithm, (Block,...), ...],

        id (str): The id to be used on the HTML component that will contain the
          results

    '''
    return dict(
            request=context['request'],
            object=object,
            results=results,
            id=id,
            URL_PREFIX=context['URL_PREFIX'],
            )


@register.assignment_tag
def block_for_experiment(xp, blocks):
    """Selects the block that belongs to the given experiment"""

    retval = [k for k in blocks if k.experiment == xp]
    if retval: return retval[0]
    return None


TO_JS_STRING=re.compile('[/_\.]')
@register.filter
def javascriptify(s):
    """Transforms a string into a name that is usable by javascript"""
    return TO_JS_STRING.sub('-', s)


@register.filter
def json_experiment_names(blocks):
    """Returns a concatenated list of experiments from the blocks"""
    return json.dumps([k.experiment.fullname() for k in blocks])


@register.filter
def json_block_names(blocks):
    """Returns a concatenated list of block names from the blocks"""
    return json.dumps([k.name for k in blocks])


@register.filter
def sortable_list(l):
    """Returns a concatenated list of block names from the blocks"""
    return (','.join(l)).replace('/','')


@register.assignment_tag
def plot_details_for_result(r):
    """Returns plotter/parameter information for the provided result"""

    # get capable plotters and parameters associated
    plotting = Plotter.objects.for_strformat(r.type)

    return dict(
            plotter=plotting['default'][0].fullname(),
            parameter=plotting['default'][1].fullname(),
            parameters=json.dumps([k.fullname() for k in plotting['options'][plotting['default'][0]]]),
            )


@register.assignment_tag
def user_subscribed_to_leaderboard(search, user):
    """Checks if a given user is subscribed to the search leaderboard"""
    if not search.has_leaderboard(): return False
    return search.leaderboard.notify.filter(id=user.id).exists()
