#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.test import TestCase
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse

from datetime import datetime
from datetime import date
from datetime import timedelta

from .models import HourlyStatistics
from .utils import updateStatistics

from ..dataformats.models import DataFormat

from ..common.testutils import BaseTestCase

import beat.core.stats

import nose.tools

STATS_1 = {
        'cpu': { 'user': 1, 'system': 1.5, },
        'memory': { 'rss': 100, },
        'data': {
            'volume': { 'read': 1000, 'write': 2000, },
            'blocks': { 'read': 100, 'write': 200, },
            'time': { 'read': 10000, 'write': 20000, },
            'generated_files': [],
            }
        }


STATS_2 = {
        'cpu': { 'user': 2, 'system': 3, },
        'memory': { 'rss': 200, },
        'data': {
            'volume': { 'read': 500, 'write': 600, },
            'blocks': { 'read': 50, 'write': 60, },
            'time': { 'read': 5000, 'write': 6000, },
            'generated_files': [],
            }
        }


class UpdateStatisticsFunction(TestCase):

    def test_first_update(self):
        statistics = beat.core.stats.Statistics(STATS_1)

        date1 = datetime(2013, 10, 5, 6, 30, 45)
        updateStatistics(statistics, date=date1)

        nose.tools.eq_(HourlyStatistics.objects.count(), 1)

        hourly = HourlyStatistics.objects.all()[0]
        nose.tools.eq_(hourly.date, date1.date())
        nose.tools.eq_(hourly.hour, 6)

        nose.tools.eq_(hourly.cpu_time                    , 2.5)
        nose.tools.eq_(hourly.max_memory                  , 100)
        nose.tools.eq_(hourly.data_read_size              , 1000)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 100)
        nose.tools.eq_(hourly.data_read_time              , 10000)
        nose.tools.eq_(hourly.data_written_size           , 2000)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 200)
        nose.tools.eq_(hourly.data_written_time           , 20000)


    def test_two_updates_in_the_same_hour(self):
        statistics = beat.core.stats.Statistics(STATS_1)

        date1 = datetime(2013, 10, 5, 6, 30, 45)
        updateStatistics(statistics, date=date1)

        statistics = beat.core.stats.Statistics(STATS_2)

        date2 = datetime(2013, 10, 5, 6, 45, 0)
        updateStatistics(statistics, date=date2)


        nose.tools.eq_(HourlyStatistics.objects.count(), 1)

        hourly = HourlyStatistics.objects.all()[0]
        nose.tools.eq_(hourly.date, date1.date())
        nose.tools.eq_(hourly.hour, 6)

        nose.tools.eq_(hourly.cpu_time                    , 7.5)
        nose.tools.eq_(hourly.max_memory                  , 300)
        nose.tools.eq_(hourly.data_read_size              , 1500)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 150)
        nose.tools.eq_(hourly.data_read_time              , 15000)
        nose.tools.eq_(hourly.data_written_size           , 2600)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 260)
        nose.tools.eq_(hourly.data_written_time           , 26000)


    def test_two_updates_in_different_hours(self):
        statistics = beat.core.stats.Statistics(STATS_1)

        date1 = datetime(2013, 10, 5, 6, 30, 45)
        updateStatistics(statistics, date=date1)

        statistics = beat.core.stats.Statistics(STATS_2)

        date2 = datetime(2013, 10, 5, 7, 45, 0)
        updateStatistics(statistics, date=date2)


        nose.tools.eq_(HourlyStatistics.objects.count(), 2)

        hourly = HourlyStatistics.objects.all()[0]
        nose.tools.eq_(hourly.date, date1.date())
        nose.tools.eq_(hourly.hour, 6)

        nose.tools.eq_(hourly.cpu_time                    , 2.5)
        nose.tools.eq_(hourly.max_memory                  , 100)
        nose.tools.eq_(hourly.data_read_size              , 1000)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 100)
        nose.tools.eq_(hourly.data_read_time              , 10000)
        nose.tools.eq_(hourly.data_written_size           , 2000)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 200)
        nose.tools.eq_(hourly.data_written_time           , 20000)


        hourly = HourlyStatistics.objects.all()[1]
        nose.tools.eq_(hourly.date, date2.date())
        nose.tools.eq_(hourly.hour, 7)

        nose.tools.eq_(hourly.cpu_time                    , 5.0)
        nose.tools.eq_(hourly.max_memory                  , 200)
        nose.tools.eq_(hourly.data_read_size              , 500)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 50)
        nose.tools.eq_(hourly.data_read_time              , 5000)
        nose.tools.eq_(hourly.data_written_size           , 600)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 60)
        nose.tools.eq_(hourly.data_written_time           , 6000)


    def test_two_updates_in_different_days(self):
        statistics = beat.core.stats.Statistics(STATS_1)

        date1 = datetime(2013, 10, 5, 6, 30, 45)
        updateStatistics(statistics, date=date1)

        statistics = beat.core.stats.Statistics(STATS_2)

        date2 = datetime(2013, 10, 6, 7, 45, 0)
        updateStatistics(statistics, date=date2)


        nose.tools.eq_(HourlyStatistics.objects.count(), 2)

        hourly = HourlyStatistics.objects.all()[0]
        nose.tools.eq_(hourly.date, date1.date())
        nose.tools.eq_(hourly.hour, 6)

        nose.tools.eq_(hourly.cpu_time                    , 2.5)
        nose.tools.eq_(hourly.max_memory                  , 100)
        nose.tools.eq_(hourly.data_read_size              , 1000)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 100)
        nose.tools.eq_(hourly.data_read_time              , 10000)
        nose.tools.eq_(hourly.data_written_size           , 2000)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 200)
        nose.tools.eq_(hourly.data_written_time           , 20000)

        hourly = HourlyStatistics.objects.all()[1]
        nose.tools.eq_(hourly.date, date2.date())
        nose.tools.eq_(hourly.hour, 7)

        nose.tools.eq_(hourly.cpu_time                    , 5.0)
        nose.tools.eq_(hourly.max_memory                  , 200)
        nose.tools.eq_(hourly.data_read_size              , 500)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 50)
        nose.tools.eq_(hourly.data_read_time              , 5000)
        nose.tools.eq_(hourly.data_written_size           , 600)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 60)
        nose.tools.eq_(hourly.data_written_time           , 6000)


    def test_two_updates_in_different_weeks(self):
        statistics = beat.core.stats.Statistics(STATS_1)

        date1 = datetime(2013, 10, 5, 6, 30, 45)
        updateStatistics(statistics, date=date1)

        statistics = beat.core.stats.Statistics(STATS_2)

        date2 = datetime(2013, 10, 7, 7, 45, 0)
        updateStatistics(statistics, date=date2)


        nose.tools.eq_(HourlyStatistics.objects.count(), 2)

        hourly = HourlyStatistics.objects.all()[0]
        nose.tools.eq_(hourly.date, date1.date())
        nose.tools.eq_(hourly.hour, 6)

        nose.tools.eq_(hourly.cpu_time                    , 2.5)
        nose.tools.eq_(hourly.max_memory                  , 100)
        nose.tools.eq_(hourly.data_read_size              , 1000)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 100)
        nose.tools.eq_(hourly.data_read_time              , 10000)
        nose.tools.eq_(hourly.data_written_size           , 2000)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 200)
        nose.tools.eq_(hourly.data_written_time           , 20000)

        hourly = HourlyStatistics.objects.all()[1]
        nose.tools.eq_(hourly.date, date2.date())
        nose.tools.eq_(hourly.hour, 7)

        nose.tools.eq_(hourly.cpu_time                    , 5.0)
        nose.tools.eq_(hourly.max_memory                  , 200)
        nose.tools.eq_(hourly.data_read_size              , 500)
        nose.tools.eq_(hourly.data_read_nb_blocks         , 50)
        nose.tools.eq_(hourly.data_read_time              , 5000)
        nose.tools.eq_(hourly.data_written_size           , 600)
        nose.tools.eq_(hourly.data_written_nb_blocks      , 60)
        nose.tools.eq_(hourly.data_written_time           , 6000)
