#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.contrib.auth.models import User

from rest_framework.reverse import reverse
from rest_framework import serializers

from .models import Team
from ..common.models import Contribution
from signals import added_to_team

#----------------------------------------------------------


class BasicTeamSerializer(serializers.ModelSerializer):
    """No name and no members"""

    name = serializers.SerializerMethodField()
    is_owner = serializers.SerializerMethodField()
    accessibility = serializers.SerializerMethodField()
    members = serializers.SerializerMethodField()

    class Meta:
        model = Team
        fields = ['short_description', 'is_owner', 'accessibility']

    def get_name(self, obj):
        return obj.fullname()

    def get_is_owner(self, obj):
        return (obj.owner == self.context.get('user'))

    def get_accessibility(self, obj):
        if obj.privacy_level == Team.PUBLIC:
            return 'public'
        elif obj.privacy_level == Team.MEMBERS:
            return 'members'
        else:
            return 'private'

    def get_members(self, obj):
        return map(lambda x: x.username, obj.members.iterator())


#----------------------------------------------------------


class SimpleTeamSerializer(BasicTeamSerializer):

    class Meta(BasicTeamSerializer.Meta):
        fields = ['name', 'short_description', 'is_owner', 'accessibility']


#----------------------------------------------------------


class FullTeamSerializer(BasicTeamSerializer):

    class Meta(BasicTeamSerializer.Meta):
        fields = ['name', 'short_description', 'is_owner', 'accessibility', 'members']


#----------------------------------------------------------


class CreatedTeamSerializer(BasicTeamSerializer):
    object_view = serializers.SerializerMethodField()

    class Meta(BasicTeamSerializer.Meta):
        fields = ['name',  'short_description', 'accessibility', 'members', 'object_view']

    def get_object_view(self, obj):
        return reverse('teams:view', args=[obj.owner.username, obj.name])


#----------------------------------------------------------


class UpdatedTeamSerializer(BasicTeamSerializer):

    class Meta(BasicTeamSerializer.Meta):
        fields = ['short_description', 'accessibility', 'members']


#----------------------------------------------------------


class TeamCreationSerializer(serializers.ModelSerializer):
    accessibility = serializers.CharField(source='privacy_level')
    members = serializers.ListField(child=serializers.CharField())

    class Meta:
        model = Team
        fields = ['name', 'short_description', 'accessibility', 'members']

    def create(self, validated_data):
        members = validated_data['members']
        del validated_data['members']

        team = None
        try:
            team = Team.objects.get(owner=self.context['request'].user, name=validated_data['name'])
        except:
            pass

        if team is not None:
            raise serializers.ValidationError('A team with this name already exists')

        team = self.Meta.model.objects.create(owner=self.context['request'].user, **validated_data)

        if members:
            for member in members:
                team.members.add(member)
            added_to_team.send(team, members=members)

        return team

    def update(self, instance, validated_data):
        instance.short_description = validated_data.get('short_description', instance.short_description)
        instance.privacy_level     = validated_data.get('privacy_level', instance.privacy_level)
        instance.save()

        members = validated_data.get('members', [])

        members_to_remove = [user for user in instance.members.all() if user not in members]
        members_to_add = [user for user in members if user not in instance.members.all()]

        for user in members_to_remove:
            instance.members.remove(user)

        if members_to_add:
            for user in members_to_add:
                instance.members.add(user)

            added_to_team.send(instance, members=members_to_add)

        return instance

    def validate_name(self, value):
        return Contribution.sanitize_name(value)

    def validate_accessibility(self, value):
        if value == 'public':
            return Team.PUBLIC
        elif value == 'members':
            return Team.MEMBERS
        else:
            return Team.PRIVATE

    def validate_members(self, value):
        return User.objects.filter(username__in=value)

    def to_representation(self, obj):
        serializer = CreatedTeamSerializer(obj)
        return serializer.data


#----------------------------------------------------------


class TeamUpdateSerializer(TeamCreationSerializer):

    class Meta(TeamCreationSerializer.Meta):
        fields = ['short_description', 'accessibility', 'members']

    def to_representation(self, obj):
        serializer = UpdatedTeamSerializer(obj)
        return serializer.data
