/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 * 
 * This file is part of the beat.web module of the BEAT platform.
 * 
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 * 
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.ui.smartselector.js

   Implementation of the smart selector panel
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.ui === undefined)
    beat.ui = {}


/********************************* CLASS: SmartSelector *********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// panel_id: ID of the DOM element to use
//----------------------------------------------------------------------------------------
beat.ui.SmartSelector = function(panel_id)
{
    // Attributes
    this.id              = panel_id;
    this.element         = $('#' + panel_id)[0];
    this.background      = null;
    this.entries         = null;
    this.selected_entry  = 0;
    this.current_search  = '';

    // Callbacks
    this.onEntrySelected = null;
    this.onCancel = null;

    // Behavior
    var panel = this;

    $(this.element).find('input.search').keydown(function(event)
        {
            // Up arrow
            if (event.which == 38)
            {
                var index = panel.selected_entry - 1;
                while (index >= 0)
                {
                    if (panel.entries[index].visible)
                        break;
                    --index;
                }

                if (index >= 0)
                {
                    $(panel.entries[panel.selected_entry].row).removeClass('highlighted');
                    $(panel.entries[index].row).addClass('highlighted');
                    panel.selected_entry = index;
                }

                return false;
            }

            // Down arrow
            else if (event.which == 40)
            {
                var index = panel.selected_entry + 1;
                while (index < panel.entries.length)
                {
                    if (panel.entries[index].visible)
                        break;
                    ++index;
                }

                if (index < panel.entries.length)
                {
                    $(panel.entries[panel.selected_entry].row).removeClass('highlighted');
                    $(panel.entries[index].row).addClass('highlighted');
                    panel.selected_entry = index;
                }

                return false;
            }

            // Enter key
            else if (event.which == 13)
            {
                if (panel.onEntrySelected)
                {
                    panel.onEntrySelected(panel.entries[panel.selected_entry].name,
                                          panel.entries[panel.selected_entry].data);
                }

                $(panel.element).hide();

                $(panel.background).remove();
                panel.background = null;
                return false;
            }

            // Escape key
            else if (event.which == 27)
            {
                if (panel.onCancel)
                    panel.onCancel();

                $(panel.element).hide();
                $(panel.background).remove();
                panel.background = null;
                return false;
            }
        }
    );

    $(this.element).find('input.search').keyup(function(event)
        {
            if (this.value == panel.current_search)
                return true;

            panel.current_search = this.value;

            var pattern = '.*';
            var characters = '';

            var search = this.value.trim().replace(/ /g, '');

            var valid_character = /(\w|\d|_|-)/i;

            for (var i = 0; i < search.length; ++i)
            {
                if (search[i] == '/')
                {
                    pattern += '\\' + search[i] + '.*';
                    characters += '/';
                }
                else if (valid_character.test(search[i]))
                {
                    pattern += search[i] + '.*';
                    characters += search[i];
                }
            }

            pattern = new RegExp(pattern, 'i');

            if (panel.selected_entry >= 0)
                $(panel.entries[panel.selected_entry].row).removeClass('highlighted');

            panel.selected_entry = -1;

            for (var i = 0; i < panel.entries.length; ++i)
            {
                var entry = panel.entries[i];

                var visible = pattern.test(entry.name);

                if (visible)
                {
                    var html = '';

                    var offset = 0;
                    for (var j = 0; j < characters.length; ++j)
                    {
                        var offset_lowercase = entry.name.indexOf(characters[j].toLowerCase(), offset);
                        var offset_uppercase = entry.name.indexOf(characters[j].toUpperCase(), offset);
                        var offset2;

                        if (offset_lowercase == -1)
                            offset2 = offset_uppercase;
                        else if (offset_uppercase == -1)
                            offset2 = offset_lowercase;
                        else
                            offset2 = Math.min(offset_lowercase, offset_uppercase);

                        html += entry.name.slice(offset, offset2);
                        html += '<span>' + entry.name[offset2] + '</span>';
                        offset = offset2 + 1;
                    }

                    html += entry.name.slice(offset);

                    $(entry.name_element).html(html);

                    if (!entry.visible)
                    {
                        $(entry.row).show();
                        entry.visible = true;
                    }

                    if (panel.selected_entry == -1)
                    {
                        panel.selected_entry = i;
                        $(entry.row).addClass('highlighted');
                    }
                }
                else
                {
                    entry.name_element.textContent = entry.name;

                    if (entry.visible)
                    {
                        $(entry.row).hide();
                        entry.visible = false;
                    }
                }
            }
        }
    );

    $(this.element).click(function() {
        $(panel.element).find('input.search').focus();
        return false;
    });
}


//----------------------------------------------------------------------------------------
// Displays the panel with the provided list of entries
//----------------------------------------------------------------------------------------
beat.ui.SmartSelector.prototype.display = function(entries, left, top)
{
    var panel = this;

    this.entries = entries;
    this.selected_entry = 0;

    $(this.element).find('input.search')[0].value = '';

    var ul = $(this.element).find('div.matches ul');
    ul.empty();
    ul = ul[0];

    for (var i = 0; i < entries.length; ++i)
    {
        var entry = this.entries[i];
        entry.visible = true;

        entry.row = document.createElement('li');
        entry.row.index = i;
        ul.appendChild(entry.row);

        if (i == 0)
            entry.row.className = 'highlighted';

        entry.name_element = document.createElement('p');
        entry.name_element.className = 'name';
        entry.name_element.textContent = entry.name;
        entry.row.appendChild(entry.name_element);

        if (entry.description)
        {
            var p = document.createElement('p');
            p.className = 'description';
            p.textContent = entry.description;
            entry.row.appendChild(p);
        }

        $(entry.row).mouseenter(function() {
            $(panel.entries[panel.selected_entry].row).removeClass('highlighted');
            $(this).addClass('highlighted');
            panel.selected_entry = this.index;
            return false;
        });

        $(entry.row).click(function() {
            if (panel.onEntrySelected)
            {
                panel.onEntrySelected(panel.entries[panel.selected_entry].name,
                                      panel.entries[panel.selected_entry].data);
            }

            $(panel.element).hide();

            $(panel.background).remove();
            panel.background = null;
            return false;
        });
    }

    var width         = $(this.element).width();
    var height        = $(this.element).height();
    var window_width  = $(window).width();
    var window_height = $(window).height();

    if (left === undefined)
    {
        left = (window_width - width) / 2;
    }
    else
    {
        left -= $(window).scrollLeft();

        if (left + width >= window_width)
            left = window_width - width - 4;
    }
        
    if (top === undefined)
    {
        top = (window_height - height) / 2;
    }
    else
    {
        top -= $(window).scrollTop();

        if (top + height >= window_height)
            top = window_height - height - 4;
    }

    this.element.style.top  = '' + top + 'px';
    this.element.style.left = '' + left + 'px';

    this.background = document.createElement('div');
    this.background.className = 'ui-widget-overlay ui-front';

    $(this.background).insertBefore(this.element);

    $(this.background).click(function() {
        $(panel.element).find('input.search').focus();
        return false;
    });

    $(this.element).show();
    $(this.element).find('input.search').focus();
}
