/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 * 
 * This file is part of the beat.web module of the BEAT platform.
 * 
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 * 
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/**
 * ui/js/widgets.js
 *
 * Implementation of common widgets to be re-used on different apps
 */


/**
 * Allows us to manage navigation tabs with history support
 *
 * Parameters:
 *
 *   ul_selector (string): A string with jQuery selector of the ul element with
 *     the tabs need me to manage. (e.g. "ul#object-tabs")
 */
function manage_tabs(ul_selector) {

  var navbox = $(ul_selector);
  var default_hash = navbox.find('.active a').attr('href');

  // activate tab on click, set all links at once
  navbox.on('click', 'a', function (e) {
    var $this = $(this);
    e.preventDefault();
    var hash = $this.attr('href');
    if (history.pushState) history.pushState(null, null, hash);
    else window.location.hash = hash;
    $this.tab('show'); //activates the clicked tab
  });

  function refreshHash() {
    var tab = undefined;
    if (window.location.hash == "") {
      tab = navbox.find('a[href="' + default_hash + '"]');
    }
    else {
      tab = navbox.find('a[href="' + window.location.hash + '"]');
    }
    navbox.find('a').blur();
    tab.show();
    tab.tab('show');
  }

  // You might need a polyfill for this event
  $(window).bind('hashchange', refreshHash);
  if(window.location.hash){
    // show right tab on load (read hash from address bar)
    refreshHash();
  }

}


/**
 * Treats the error condition from the platform or alerts the user generically
 *
 * Parameters:
 *
 *   data (Array): The array that contains the answer from the server
 *   status_text (str): The same as data.textStatus
 *   redirect (str): An (optional) url to redirect the client in case of an
 *     error.
 */
function process_error(data, status_text, redirect) {

  if (data.status >= 400 && data.status < 500) { //controlled response from the platform
    var response = data.responseJSON;

    var div = document.createElement('div');
    var p = document.createElement('h5');

    p.textContent = 'Error:';
    div.appendChild(p);
    var ul = document.createElement('ul');
    div.appendChild(ul);

    if (Array.isArray(response)) {
      var length = response.length;
      for (i=0; i<length; ++i) {
        var li = document.createElement('li');
        ul.appendChild(li);
        li.textContent = response[i];
      }
    }
    else if (typeof response === 'string') {
      p.textContent += ' ' + response;
    }
    else { //object with specific details on fields
      for (var key in response) {
        var li = document.createElement('li');
        ul.appendChild(li);
        li.textContent = 'Field "' + key + '":';
        if (Array.isArray(response[key])) {
          var length = response[key].length;
          var sub_ul = document.createElement('ul');
          li.appendChild(sub_ul);
          for (i=0; i<length; ++i) {
            var sub_li = document.createElement('li');
            sub_li.textContent = response[key][i];
            sub_ul.appendChild(sub_li);
          }
        }
        else { //assumed to be String
          li.textContent += ' ' + response[key];
        }
      }
    }

    BootstrapDialog.alert({
      type: BootstrapDialog.TYPE_DANGER,
      title: 'Error during submission' + ' - ' + data.statusText + ' (' + data.status + ')',
      message: div,
      onhidden: function() {
        if (redirect !== undefined) window.location = redirect;
      }
    });
  }

  else if (data.status == 0) { //communication error
    BootstrapDialog.alert({
      message: 'An empty response was received by your browser, for a request you sent. This normally indicates it cannot establish a connection with the web service. In the hopes it is a temporary issue, you may re-submit your request again at a later moment. Otherwise, please contact the system administrators.',
      type: BootstrapDialog.TYPE_DANGER,
      title: 'Empty response',
      onhidden: function() {
        if (redirect !== undefined) window.location = redirect;
      }
    });
  }
  else { //unknown error
    BootstrapDialog.alert({
      message: data.responseJSON,
      type: BootstrapDialog.TYPE_DANGER,
      title: data.statusText + ' (' + data.status + ')',
      onhidden: function() {
        if (redirect !== undefined) window.location = redirect;
      }
    });
  }

}


/**
 * Creates an alert box and returns it w/o attaching it to the DOM
 *
 * Parameters:
 *
 *   html (String): The text that will be placed inside the alert box
 *   dismissible (Boolean): If the alert is supposed to be dismissible or not
 *   type (String): One of "success", "info" "warning", "danger"
 */
function make_alert(text, dismissible, type) {
  var retval = $(document.createElement('div')).addClass('alert alert-' + type);
  retval.attr('role', 'alert');
  if (dismissible) {
    retval.addClass('alert-dismissible');
    var btn = $(document.createElement('button')).addClass('close');
    btn.attr('type', 'button');
    btn.attr('data-dismiss', 'alert');
    btn.attr('aria-label', 'Close');
    btn.append($(document.createElement('span')).attr('aria-hidden', 'true').html('&times;'));
    retval.append(btn);
  }
  retval.append($(text));
  return retval;
}


/**
 * Uses bootstrap3 to display a modal window to confirm the submission of a
 * form.
 *
 * If the user confirms the submission, it is carried out as if this modal did
 * not show up. Redirections and other error handling is to be taken care by
 * the server.
 *
 * Parameters:
 *
 * event (Object): The jQuery event being halted so this modal appears
 *   (normally, the *   form submission itself.
 * form (Object): The jQuery form element that was being submitted
 * title (str, optional): If set, will be set on the title bar of the modal
 *   window. If not set, then the modal window will have no title bar.
 * message (str, optional): The message to be displayed to the user on the
 *   modal window. If not set, no message field will appear on the window.
 **/
function confirm_form_submission(event, form, title, message) {
  event.preventDefault();
  var dialog = new BootstrapDialog({
    title: title,
    message: message,
    type: BootstrapDialog.TYPE_PRIMARY,
    buttons: [
    {
      label: 'Cancel',
      cssClass: 'btn-default',
      action: function(the_dialog) {
        the_dialog.close();
      }
    },
    {
      label: 'Continue',
      cssClass: 'btn-primary',
      action: function(the_dialog) {
        form.submit();
        the_dialog.close();
      }
    },
    ],
    });
  dialog.realize();
  if (title === undefined) dialog.getModalHeader().hide();
  if (message === undefined) dialog.getModalBody().hide();
  dialog.open();
}


/**
 * Uses bootstrap3 to display a modal window to delete a contribution
 *
 * This method will display an alert box if a problem occurs with the deletion,
 * or will just redirect the client to the `redirect` URL in case of success.
 *
 * Parameters:
 *
 * type (str): The type of contribution to delete (e.g. "toolchain")
 * name (str): The full name of the contribution to delete (e.g.
 *   "user/toolchain/1")
 * url (str): The URL where to post the DELETE query (e.g.
 *   "/api/v1/toolchains/user/toolchain/1/")
 * redirect (str): The URL where to redirect the client after the operation
 *   finished (e.g. "/toolchains/user/")
 **/
function modal_delete(type, name, url, redirect) {

  var dialog = new BootstrapDialog({
    title: '<i class="fa fa-warning"></i> Confirm deletion of ' +
      type + ' "' + name + '"',
    message: 'Once you delete this ' + type +
      ', there is no going back. Be certain.',
    type: BootstrapDialog.TYPE_PRIMARY,
    buttons: [
    {
      label: 'Cancel',
      cssClass: 'btn-default',
      action: function(the_dialog) {
        the_dialog.close();
      }
    },
    {
      label: 'Delete',
      cssClass: 'btn-primary',
      action: function(the_dialog) {
        $.ajaxSetup({
          beforeSend: function(xhr, settings) {
            var csrftoken = $.cookie('csrftoken');
            xhr.setRequestHeader('X-CSRFToken', csrftoken);
          }
        });
        $.ajax({
          type: "DELETE",
          url: url + name + '/',
          success: function(data) {
            the_dialog.close();
            window.location.href = redirect;
          },

          error: function(jqXHR, textStatus, errorThrown) {
            if ((jqXHR.status == 403) && (jqXHR.responseText.length > 0))
              alert(jqXHR.responseText);
            else
              alert('Error: ' + errorThrown);
            the_dialog.close();
          }
        });
        the_dialog.close();
      }
    },
    ],
    });
  dialog.realize();
  dialog.open();

}

/**
 * Checks the name of a contribution using the provided input text box and the
 * checking URL.
 *
 * Toggles error using "has-error" or success as "has-success" and replaces the
 * name on the input box with the one provided by the API.
 */
function check_name (group_id, input_id, url) {

  var group = $('#'+group_id);
  var text_field = $('#'+input_id);
  if (text_field.length == 0) return false; //does not exist, ignore
  var value = text_field.val();
  if (value.length == 0) return false; //it is empty, ignore

  $.ajaxSetup({
    beforeSend: function(xhr, settings) {
      var csrftoken = $.cookie('csrftoken');
      xhr.setRequestHeader('X-CSRFToken', csrftoken);
    }
  });

  $.ajax({
    type: 'POST',
    url: url,
    data: JSON.stringify({name: value}),
    contentType: "application/json; charset=utf-8",
    dataType: "json",
    success: function(data){
      text_field.val(data.name);
      group.toggleClass('has-success', !data.used);
      group.toggleClass('has-error', data.used);
      if (data.used) {
        BootstrapDialog.alert({
          title: 'Name clash',
          message: "The name <b>" + data.name + "</b> is already in use. Choose another one",
          type: BootstrapDialog.TYPE_DANGER,
        });
      }
    },
    error: function(data, textStatus){
      BootstrapDialog.alert({
        title: textStatus + ' (' + data.status + ')',
        message: data.responseText,
        type: BootstrapDialog.TYPE_DANGER,
      });
    },
  });
}

/**
 * Captures the "uses" dictionary from a prepare table in the edition of an
 * algorithm or library.
 *
 * Checks for empty import names, repeated import names or unselected library
 * names.
 */
function get_uses (row_selector, val_selector) {
  var retval = {};
  $(row_selector).each(function() {
    var key = $(this).find(val_selector).val();
    var val = $(this).find('select').val();
    $(this).removeClass('has-success has-error success danger');
    if (key.length == 0) {
      BootstrapDialog.alert({
        title: 'Empty import name',
        message: "Set proper import names for all rows in your used libraries table",
        type: BootstrapDialog.TYPE_DANGER,
      });
      $(this).addClass('has-error danger');
      retval = false;
      return retval;
    }
    else if (key in retval) {
      BootstrapDialog.alert({
        title: 'Duplicated import name',
        message: "You cannot import two libraries with the same import name. Choose a unique name for each import in your used libraries table",
        type: BootstrapDialog.TYPE_DANGER,
      });
      $(this).addClass('has-error danger');
      retval = false;
      return retval;
    }
    else if (!val) {
      BootstrapDialog.alert({
        title: 'No library selected',
        message: "Make sure you select one existing library for every import row you have on your used libraries table",
        type: BootstrapDialog.TYPE_DANGER,
      });
      $(this).addClass('has-error danger');
      retval = false;
      return retval;
    }
    retval[key] = val;
  });
  if (retval == false) return retval;
  $(this).addClass('has-success success');
  return retval;
}

/**
 * Expands headers of objects in detriment of the action buttons
 *
 * This function assumes the following organization of the header:
 *
 * <div><!-- the header for this view page -->
 *   <div class="col-sm-9"
 *        onmouseover="expand_breadcrumb(this, 9, 3, 12);"
 *        onmouseout="reset_breadcrumb(this, 9, 3, 12);">
 *     <!-- the name of the object -->
 *   </div>
 *   <div class="col-sm-3">
 *     <!-- the actions for the object -->
 *   </div>
 * </div>
 *
 * If deployed this way, it will expand the associated diff to the size
 * specified by `s3`. If `s3` is not specified, it uses the value `12`. The
 * values for `s1` and `s2` define the current sizes for the first and the
 * second `<div>` objects.
 */
function expand_breadcrumb(div, s1, s2, s3) {
  s3 = typeof s3 !== 'undefined' ? s3 : 12;
  $(div).removeClass('col-sm-' + s1).addClass('col-sm-' + s3);
  $(div).next().removeClass('col-sm-' + s2).hide();
}

/**
 * Shrinks headers back to original size. This is the inverse function of
 * `expand_breadcrumb()`, supposed to be applied with the trigger `onmouseout`.
 * See the documentation for * `expand_breadcrumb()` above for usage details.
 */
function reset_breadcrumb(div, s1, s2, s3) {
  s3 = typeof s3 !== 'undefined' ? s3 : 12;
  $(div).removeClass('col-sm-' + s3).addClass('col-sm-' + s1);
  $(div).next().addClass('col-sm-' + s2).show();
}

/**
 * Hits an URL and reports if errors occur.
 */
function modal_email(url, author) {

  var dialog = new BootstrapDialog({
    title: '<i class="fa fa-envelope"></i> Confirm  e-mail to ' + author,
    message: "We're about to send an e-mail requesting " + author +
      ' to document this resource. The field <code>reply-to</code> of ' +
      ' this e-mail will be set to your e-mail address, so you can be ' +
      'contacted by the author when done. Do you wish to send or cancel ' +
      'this e-mail?',
    type: BootstrapDialog.TYPE_PRIMARY,
    buttons: [
    {
      label: 'Cancel',
      cssClass: 'btn-default',
      action: function(the_dialog) {
        the_dialog.close();
      }
    },
    {
      label: 'Send',
      cssClass: 'btn-primary',
      action: function(the_dialog) {
        var button = $(this);
        button.disable();
        button.html('<i class="fa fa-spin fa-refresh"></i> Sending...');
        $.ajaxSetup({
          beforeSend: function(xhr, settings) {
            var csrftoken = $.cookie('csrftoken');
            xhr.setRequestHeader('X-CSRFToken', csrftoken);
          }
        });
        $.ajax({
          type: "GET",
          url: url,
          success: function(data) {
            button.html('<i class="fa fa-check"></i> Sent!');
            button.removeClass('btn-primary').addClass('btn-success');
            setTimeout(function() { the_dialog.close(); }, 1000);
          },
          error: function(jqXHR, textStatus, errorThrown) {
            if ((jqXHR.status == 403) && (jqXHR.responseText.length > 0))
              alert(jqXHR.responseText);
            else
              alert('Error: ' + errorThrown);
            the_dialog.close();
          }
        });
      }
    },
    ],
    });
  dialog.realize();
  dialog.open();

}
