#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

# Code copied on 24.09.2014 (and customised) from:
# https://github.com/lambdalisue/django-codemirror-widget
# (originally BSD-Licensed)

from itertools import chain
import re
import hashlib
import json

from django import forms
from django.conf import settings
from django.utils.safestring import mark_safe
from django.utils import six
from .. import __version__

# set default settings
CODEMIRROR_PATH = getattr(settings, 'CODEMIRROR_PATH', 'codemirror')
if CODEMIRROR_PATH.endswith('/'):
    CODEMIRROR_PATH = CODEMIRROR_PATH[:-1]
CODEMIRROR_MODE = getattr(settings, 'CODEMIRROR_MODE', 'javascript')
CODEMIRROR_THEME = getattr(settings, 'CODEMIRROR_THEME', 'default')
CODEMIRROR_SIZE = getattr(settings, 'CODEMIRROR_SIZE', None)

THEME_CSS_FILENAME_RE = re.compile(r'[\w-]+')

def uniq(seq, idfun=None):
    """Order preserving, fast de-duplication for lists"""

    if idfun is None:
        def idfun(x): return x

    seen = {}
    result = []

    for item in seq:
        marker = idfun(item)
        if marker in seen: continue
        seen[marker] = 1
        result.append(item)
    return result

def isstring(obj):
    try:
        return isinstance(obj, basestring)
    except NameError:
        return isinstance(obj, str)


class CodeMirrorJavascript(object):
    """An object used to mark Javascript sections in JSON output.

    If an object of this type is passed to CodeMirrorJSONEncoder().encode(),
    the text containined in it will be output with no transformations applied.
    Most likely this will not be valid JSON, but it will be valid Javascript,
    assuming valid Javascript was passed to the constructor.

    Example usage:

    my_data = {
        'someKey': True,
        'someCallback': CodeMirrorJavascript("function() { return true; }")
    }

    CodeMirrorJSONEncoder().encode(my_data)

    # -> '{"someKey": true, "someCallback": function() { return true; }}'
    """

    def __init__(self, js):
        "js is a string containing valid Javascript code."
        self.js = js


class CodeMirrorJSONEncoder(json.JSONEncoder):
    """A custom JSON encoder that knows how to handle CodeMirrorJavascript()
    objects."""

    stash_prefix = "js_stash::"

    def __init__(self, *args, **kwargs):
        super(CodeMirrorJSONEncoder, self).__init__(*args, **kwargs)
        self.stash = {}

    def default(self, obj):
        if isinstance(obj, CodeMirrorJavascript):
            # If a Javascript object is encountered, replace it with a placeholder.
            stash_id = self.stash_prefix + hashlib.md5(obj.js).hexdigest()
            self.stash[stash_id] = obj.js
            return stash_id
        return super(CodeMirrorJSONEncoder, self).default(obj)

    def encode(self, obj):
        self.stash = {}
        encoded = super(CodeMirrorJSONEncoder, self).encode(obj)
        # Search for any placeholders and replace them with their original values.
        for key, val in self.stash.items():
            encoded = encoded.replace('"' + key + '"', val)
        self.stash = {}
        return encoded

class CodeMirrorTextarea(forms.Textarea):
    u"""Textarea widget render with `CodeMirror`

    CodeMirror: http://codemirror.net/
    """

    @property
    def media(self):
        mode_name = self.mode_name
        js = [
                "%s/lib/codemirror.js" % (CODEMIRROR_PATH,),
                "%s/addon/fold/foldcode.js" % (CODEMIRROR_PATH,),
                "%s/addon/fold/foldgutter.js" % (CODEMIRROR_PATH,),
                ]

        if not self.custom_mode:
            js.append("%s/mode/%s/%s.js" % (CODEMIRROR_PATH, mode_name, mode_name))

        js += ["%s/mode/%s/%s.js" % (CODEMIRROR_PATH, dependency, dependency) for dependency in self.dependencies]
        js += ["%s/addon/%s.js" % (CODEMIRROR_PATH, addon) for addon in self.addon_js]

        if self.keymap:
            js.append("%s/keymap/%s.js" % (CODEMIRROR_PATH, self.keymap))

        if self.custom_js:
            js.extend(self.custom_js)

        # by the end, add the defaults
        js += ["ui/js/codemirror-defaults.js"]

        css = (
                "%s/lib/codemirror.css" % CODEMIRROR_PATH,
                ) + \
              tuple("%s/theme/%s.css" % (CODEMIRROR_PATH, theme_css_filename)
                      for theme_css_filename in self.theme_css) + \
              tuple("%s/addon/%s.css" % (CODEMIRROR_PATH, css_file)
                      for css_file in self.addon_css) + \
              (
                "ui/css/codemirror-overrides.css",
                "%s/addon/fold/foldgutter.css" % CODEMIRROR_PATH,
                )

        # adds version fingerprinting to javascript/css files
        js = ['%s?v=%s' % (k, __version__) for k in js]
        css = ['%s?v=%s' % (k, __version__) for k in css]

        return forms.Media(css={'all': css},
            js=js
        )

    def __init__(
            self, attrs=None, mode=None, theme=None, config=None, size=None,
            dependencies=(), addon_js=(), addon_css=(),
            custom_mode=None, custom_js=(), keymap=None, **kwargs):
        u"""Constructor of CodeMirrorTextarea

        Attribute:
            path          - CodeMirror directory URI (DEFAULT =
                            settings.CODEMIRROR_PATH)
            mode          - Name of language or a modal configuration object as
                            described in CodeMirror docs.  Used to autoload an
                            appropriate language plugin js file according to
                            filename conventions.  (DEFAULT =
                            settings.CODEMIRROR_MODE)
            theme         - Name of theme. Also autoloads theme plugin css
                            according to filename conventions.
                            (DEFAULT = settings.CODEMIRROR_THEME)
            config        - The rest of the options passed into CodeMirror as a
                            python map.
            size          - Size of the editable area as a tuple
                            (width, height) (updated from
                            settings.CODEMIRROR_SIZE)
            dependencies  - Some modes depend on others, you can pass extra
                            modes dependencies with this argument. For example
                            for mode="htmlmixed", you must pass
                            dependencies=("xml", "javascript", "css").
            addon_js      - Various addons are available for CodeMirror. You
                            can pass the names of any addons to load with this
                            argument. For example, for mode="django", you must
                            pass addons=("mode/overlay", ).
            addon_css     - Some addons require corresponding CSS files. Since
                            not every addon requires a CSS file, and the names
                            of these files do not always follow a convention,
                            they must be listed separately. For example,
                            addon_css=("display/fullscreen", ).
            custom_mode   - To use a custom mode (i.e. one not included in the
                            standard CodeMirror distribution), set this to
                            the name, or configuration object, of the mode, and
                            ensure "mode" is None. For example,
                            custom_mode="my_custom_mode".
            custom_js     - To include other Javascript files with this widget
                            that are not defined in the CodeMirror package, set
                            this to a list of pathnames. If "custom_mode" is
                            defined, this will probably contain the path of the
                            file defining that mode. Paths in this list will
                            not be prepended with settings.CODEMIRROR_PATH.
                            For example,
                            custom_js=("site_js/my_custom_mode.js", )
            keymap        - The name of a keymap to use. Keymaps are located in
                             settings.CODEMIRROR_PATH/keymap. Default: None.

        Example:
            *-------------------------------*
            + static
              + codemirror
                + lib
                  - codemirror.js
                  - codemirror.css
                + mode
                  + python
                    - python.js
                + theme
                  + cobalt.css
                + addon
                  + display
                    - fullscreen.js
                    - fullscreen.css
              + site_js
                - my_custom_mode.js
            *-------------------------------*
            CODEMIRROR_PATH = "codemirror"

            codemirror = CodeMirrorTextarea(mode="python", theme="cobalt", config={ 'fixedGutter': True })
            document = forms.TextField(widget=codemirror)
        """

        super(CodeMirrorTextarea, self).__init__(attrs=attrs, **kwargs)

        self.mode = mode or custom_mode or CODEMIRROR_MODE
        if isstring(self.mode):
            self.mode = { 'name': self.mode }
        self.mode_name = self.mode['name']
        self.size = size or CODEMIRROR_SIZE
        self.custom_mode = custom_mode
        self.dependencies = dependencies
        self.addon_js = addon_js
        self.addon_css = addon_css
        self.custom_js = custom_js
        self.keymap = keymap

        self.addon_js = uniq(list(addon_js) + [
                "mode/overlay",
                "search/search",
                "search/searchcursor",
                "dialog/dialog",
                ])

        self.addon_css = uniq(list(addon_css) + ["dialog/dialog"])

        self.theme = theme or CODEMIRROR_THEME
        theme_css_filename = THEME_CSS_FILENAME_RE.search(self.theme).group(0)
        if theme_css_filename == 'default':
            self.theme_css = []
        else:
            self.theme_css = [theme_css_filename]

        self.config = config or {}

    def render(self, name, value, attrs=None):
        """Render CodeMirrorTextarea"""

        if self.size is not None:
            suffix = ' %s_editor.setSize%s;' % (name, tuple(self.size))
        else:
            suffix = ''

        self.config.setdefault('foldGutter', True)
        option_json = CodeMirrorJSONEncoder().encode(dict(chain(
            self.config.items(),
            [('mode', self.mode), ('theme', self.theme)])))

        output = [
            super(CodeMirrorTextarea, self).render(name, value, attrs),
            '<script type="text/javascript">var %(name)s_editor = CodeMirror.fromTextArea(document.getElementById("id_%(name)s"), %(options)s);%(suffix)s</script>' % \
                dict(
                  name=name,
                  options=option_json,
                  suffix=suffix,
                  ),
            ]
        return mark_safe('\n'.join(output))

class CodeMirrorFileWidget(CodeMirrorTextarea):
    """Allows editing a file instead of a plain content"""

    def __init__(self, *args, **kwargs):
        super(CodeMirrorFileWidget, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs=None):
        if isinstance(value, six.string_types):
            contents = value
        elif hasattr(value, 'read'):
            contents = value.read()
        else:
            contents = ''
        return mark_safe(super(CodeMirrorFileWidget, self).render(name, contents, attrs))
