#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import logging
logger = logging.getLogger(__name__)

import os
import time
import copy
import shutil
import tarfile
import tempfile
import datetime

from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.conf import settings
from django.apps import apps

from ....import __version__

APPS = [ #dump and load order are respected
        'authtoken',
        'backend',
        'statistics',
        'dataformats',
        'databases',
        'libraries',
        'algorithms',
        'plotters',
        'toolchains',
        'experiments',
        'attestations',
        'search',
        'reports',
        'actstream',
        ]


def _check(app, queryset):
    """Checks the queryset for dependence errors, fixes problems"""

    global logger

    declared = []
    errors = 0

    for k in queryset.iterator():

        # checks the forks
        if k.fork_of and k.fork_of.id not in declared:
            logger.warn("** Inconsistent creation date - %s `%s' (%s) which is a fork of `%s' (%s) must be dumped after, but was created before", app, k, k.creation_date, k.fork_of, k.fork_of.creation_date)
            logger.warn("   -> Correcting creation date of %s to be (a second after) %s", k, k.fork_of)
            k.creation_date = k.fork_of.creation_date + datetime.timedelta(seconds=1)
            k.save()
            errors += 1

        # checks previous versions
        if k.previous_version and k.previous_version.id not in declared:
            logger.warn("** Inconsistent creation date - %s `%s' (%s) which has `%s' (%s) as a previous version must be dumped after, but was created before", app, k, k.creation_date, k.previous_version, k.previous_version.creation_date)
            logger.warn("   -> Correcting creation date of %s to be (a second after) %s", k, k.previous_version)
            k.creation_date = k.previous_version.creation_date + datetime.timedelta(seconds=1)
            k.save()
            errors += 1

        declared.append(k.id)


    return errors


class Command(BaseCommand):

    help = 'Backs-up current database and prefix in a single tarball'


    def handle(self, *ignored, **arguments):

        # Setup this command's logging level
        global logger
        arguments['verbosity'] = int(arguments['verbosity'])
        if arguments['verbosity'] >= 1:
            if arguments['verbosity'] == 1: logger.setLevel(logging.INFO)
            elif arguments['verbosity'] >= 2: logger.setLevel(logging.DEBUG)

        # for these apps, only backs-up these particular object types
        only = dict(
                dataformats = 'dataformats.DataFormat',
                libraries = 'libraries.Library',
                algorithms = 'algorithms.Algorithm',
                databases = 'databases.Database',
                toolchains = 'toolchains.Toolchain',
                )

        dump_arguments = dict(
                indent = 2,
                verbosity=arguments.get('verbosity'),
                interactive=False,
                use_natural_primary_keys = True,
                use_natural_foreign_keys = True,
                format = 'json',
                exclude = [
                    'sessions',
                    'admin',
                    'contenttypes',
                    'auth.Permission',
                    ],
                )

        try:
            tmpdir = tempfile.mkdtemp('.backup', 'beat.web-')

            # backs-up everything else first
            arguments = copy.deepcopy(dump_arguments)
            arguments['exclude'] += APPS
            destfile = os.path.join(tmpdir, 'initial.json')
            logger.info("Dumping initial (unspecified) data -> `%s'", destfile)
            arguments['output'] = destfile #new in Django-1.8.x
            call_command('xdumpdata', **arguments)

            # and backs-up the apps respecting the imposed order
            for app in APPS:

                destfile = os.path.join(tmpdir, '%s.json' % app)
                arguments = copy.deepcopy(dump_arguments)
                logger.info("Dumping data for `%s' -> `%s'", app, destfile)

                if app in only:

                    app, model = only[app].split('.')
                    model = apps.get_model(app, model)
                    order = ('creation_date',)

                    # This will check and correct objects with weird creation
                    # dates so that the dump order is consistent
                    while True:
                        queryset = model.objects.order_by(*order)
                        err = _check(app, queryset)
                        if not err: break

                    arguments['primary_keys'] = ','.join([str(k.id) for k in queryset])

                arguments['output'] = destfile #new in Django-1.8.x
                call_command('xdumpdata', only.get(app, app), **arguments)

                # copy prefix data
                path = os.path.join(settings.PREFIX, app)
                if os.path.exists(path):
                    destdir = os.path.join(tmpdir, 'prefix', app)
                    logger.info("Backing up core objects for `%s' -> `%s'",
                            app, destdir)
                    shutil.copytree(path, destdir)
                else:
                    logger.info("No disk presence found for `%s'", app)

            # tarball and remove directory
            compress = 'bz2'
            tarball = time.strftime('%Y.%m.%d-%Hh%Mm%S') + \
                ('-v%s' % __version__) + '.tar.' + compress
            logger.info("Writing archive `%s'", tarball)
            with tarfile.open(tarball, "w:%s" % compress) as tar:
                tar.add(tmpdir, arcname='')

        finally:

            # removes the temporary directory
            shutil.rmtree(tmpdir)
